import numpy as np
import pytest
import shapely.geometry as sg

import gdsfactory as gf

DEMO_PORT_ANGLE = 10


def assert_polygon_equals(coords_expected, coords_actual):
    coords_expected = gf.snap.snap_to_grid(coords_expected)
    coords_actual = gf.snap.snap_to_grid(coords_actual)
    shape_expected = sg.polygon.orient(sg.Polygon(coords_expected))
    shape_actual = sg.polygon.orient(sg.Polygon(coords_actual))

    assert shape_expected.equals(
        shape_actual
    ), f"Expected: {shape_expected}. Got: {shape_actual}"


def get_expected_shear_shape(length, width, shear_angle):
    dx = np.round(np.tan(np.deg2rad(shear_angle)) * width * 0.5, 3)
    x0 = 0
    x1 = length
    y0 = -width * 0.5
    y1 = width * 0.5
    return np.array(
        [
            [x0 - dx, y1],
            [x1 - dx, y1],
            [x1 + dx, y0],
            [x0 + dx, y0],
        ]
    )


@pytest.fixture
def shear_waveguide_symmetric():
    P = gf.path.straight(length=10)
    return gf.path.extrude(
        P, "strip", shear_angle_start=DEMO_PORT_ANGLE, shear_angle_end=DEMO_PORT_ANGLE
    )


@pytest.fixture
def shear_waveguide_start():
    P = gf.path.straight(length=10)
    return gf.path.extrude(
        P, "strip", shear_angle_start=DEMO_PORT_ANGLE, shear_angle_end=None
    )


@pytest.fixture
def shear_waveguide_end():
    P = gf.path.straight(length=10)
    return gf.path.extrude(
        P, "strip", shear_angle_start=None, shear_angle_end=DEMO_PORT_ANGLE
    )


@pytest.fixture
def regular_waveguide():
    P = gf.path.straight(length=10)
    return gf.path.extrude(P, "strip")


@pytest.fixture
def more_slanted_than_wide():
    P = gf.path.straight(length=0.1)
    return gf.path.extrude(P, "strip", shear_angle_start=60, shear_angle_end=60)


@pytest.fixture
def skinny():
    P = gf.path.straight(length=0.1)
    return gf.path.extrude(P, "strip")


@pytest.fixture
def test_mate_on_shear_xor_empty(
    regular_waveguide, shear_waveguide_start, shear_waveguide_end
) -> None:
    # two sheared components joined at the sheared port should appear the same as two straight component joined
    two_straights = gf.Component()
    c1 = two_straights << regular_waveguide
    c2 = two_straights << regular_waveguide
    c2.connect("o1", c1.ports["o2"])

    two_shears = gf.Component()
    c1 = two_shears << shear_waveguide_end
    c2 = two_shears << shear_waveguide_start
    c2.connect("o1", c1.ports["o2"])

    xor = gf.geometry.xor_diff(two_straights, two_shears)
    assert not xor.layers


def test_rotations_are_normal(
    regular_waveguide, shear_waveguide_start, shear_waveguide_end
) -> None:
    two_shears = gf.Component()
    c1 = two_shears << shear_waveguide_end
    c2 = two_shears << shear_waveguide_start
    c2.connect("o1", c1.ports["o2"])

    assert c2.rotation % 90 == 0


def test_area_stays_same(
    regular_waveguide,
    shear_waveguide_start,
    shear_waveguide_end,
    shear_waveguide_symmetric,
) -> None:
    components = [
        regular_waveguide,
        shear_waveguide_start,
        shear_waveguide_end,
        shear_waveguide_symmetric,
    ]
    areas = [c.area() for c in components]
    np.testing.assert_allclose(areas, desired=areas[0])


def test_area_stays_same_skinny(
    skinny,
    more_slanted_than_wide,
) -> None:
    components = [
        skinny,
        more_slanted_than_wide,
    ]
    areas = [c.area() for c in components]
    np.testing.assert_allclose(areas, desired=areas[0])


def test_mate_on_shear_xor_empty_transition() -> None:
    """two sheared components joined at the sheared port should appear the same
    as two straight component joined."""
    P = gf.path.straight(length=10)

    s = gf.Section(width=3, offset=0, layer=gf.LAYER.SLAB90, name="slab")
    X1 = gf.CrossSection(
        width=1,
        offset=0,
        layer=gf.LAYER.WG,
        name="core",
        port_names=("o1", "o2"),
        sections=[s],
    )
    s2 = gf.Section(width=2, offset=0, layer=gf.LAYER.SLAB90, name="slab")
    X2 = gf.CrossSection(
        width=0.5,
        offset=0,
        layer=gf.LAYER.WG,
        name="core",
        port_names=("o1", "o2"),
        sections=[s2],
    )
    t = gf.path.transition(X1, X2, width_type="linear")
    linear_taper = gf.path.extrude(P, t)

    P = gf.path.straight(length=10)

    s = gf.Section(width=3, offset=0, layer=gf.LAYER.SLAB90, name="slab")
    x1 = gf.CrossSection(
        width=1,
        offset=0,
        layer=gf.LAYER.WG,
        name="core",
        port_names=("o1", "o2"),
        sections=[s],
    )
    s2 = gf.Section(width=2, offset=0, layer=gf.LAYER.SLAB90, name="slab")
    x2 = gf.CrossSection(
        width=0.5,
        offset=0,
        layer=gf.LAYER.WG,
        name="core",
        port_names=("o1", "o2"),
        sections=[s2],
    )
    t = gf.path.transition(x1, x2, width_type="linear")
    linear_taper_sheared = gf.path.extrude(
        P, t, shear_angle_start=10, shear_angle_end=None
    )

    two_straights = gf.Component()
    c1 = two_straights << linear_taper
    c2 = two_straights << linear_taper
    c2.connect("o1", c1.ports["o1"])

    two_shears = gf.Component()
    c1 = two_shears << linear_taper_sheared
    c2 = two_shears << linear_taper_sheared
    c2.connect("o1", c1.ports["o1"])

    xor = gf.geometry.xor_diff(two_straights, two_shears)
    # xor.show()
    # two_straights.show()
    # two_shears.show()

    area = xor.area()
    assert area < 0.1, area


def test_mate_on_shear_xor_empty_curve() -> None:
    """two sheared components joined at the sheared port should appear the same
    as two straight component joined."""
    P = gf.path.euler()
    curve = gf.path.extrude(P, "strip")

    angle = 15
    P = gf.path.euler()
    curve_sheared1 = gf.path.extrude(P, "strip", shear_angle_end=angle)
    curve_sheared2 = gf.path.extrude(P, "strip", shear_angle_start=angle)

    two_straights = gf.Component()
    c1 = two_straights << curve
    c2 = two_straights << curve
    c2.connect("o1", c1.ports["o2"])

    two_shears = gf.Component()
    c1 = two_shears << curve_sheared1
    c2 = two_shears << curve_sheared2
    c2.connect("o1", c1.ports["o2"])

    xor = gf.geometry.xor_diff(two_straights, two_shears, precision=1e-2)
    assert not xor.layers, f"{xor.layers}"


def test_shear_angle_annotated_on_ports(
    shear_waveguide_start, shear_waveguide_end
) -> None:
    assert shear_waveguide_start.ports["o1"].shear_angle == DEMO_PORT_ANGLE
    assert shear_waveguide_start.ports["o2"].shear_angle is None

    assert shear_waveguide_end.ports["o2"].shear_angle == DEMO_PORT_ANGLE
    assert shear_waveguide_end.ports["o1"].shear_angle is None


def test_port_attributes(regular_waveguide, shear_waveguide_symmetric) -> None:
    regular_ports = [p.to_dict() for p in regular_waveguide.ports.values()]
    shear_ports = [p.to_dict() for p in shear_waveguide_symmetric.ports.values()]

    for p in shear_ports:
        shear_angle = p.pop("shear_angle")
        assert shear_angle == DEMO_PORT_ANGLE

    for p1, p2 in zip(regular_ports, shear_ports):
        for k in p.keys():
            assert p1[k] == p2[k], f"{k} differs! {p1[k]} != {p2[k]}"


def test_points_are_correct(shear_waveguide_symmetric):
    shear_angle = DEMO_PORT_ANGLE
    cs = gf.get_cross_section("strip")
    wg_width = cs.width
    length = 10
    points_expected = get_expected_shear_shape(
        length=length, width=wg_width, shear_angle=shear_angle
    )

    poly_actual = shear_waveguide_symmetric.get_polygons(by_spec=(1, 0))[0]
    assert_polygon_equals(points_expected, poly_actual)


def test_points_are_correct_wide():
    wg_width = 40
    length = 10
    P = gf.path.straight(length=length)
    shear_waveguide_symmetric = gf.path.extrude(
        P,
        {"cross_section": "strip", "settings": {"width": wg_width}},
        shear_angle_start=DEMO_PORT_ANGLE,
        shear_angle_end=DEMO_PORT_ANGLE,
    )

    shear_angle = DEMO_PORT_ANGLE

    points_expected = get_expected_shear_shape(
        length=length, width=wg_width, shear_angle=shear_angle
    )
    poly_actual = shear_waveguide_symmetric.get_polygons(by_spec=(1, 0))[0]
    assert_polygon_equals(points_expected, poly_actual)


def test_points_are_correct_short():
    wg_width = 40
    length = 0.5
    P = gf.path.straight(length=length)
    shear_waveguide_symmetric = gf.path.extrude(
        P,
        {"cross_section": "strip", "settings": {"width": wg_width}},
        shear_angle_start=DEMO_PORT_ANGLE,
        shear_angle_end=DEMO_PORT_ANGLE,
    )

    shear_angle = DEMO_PORT_ANGLE

    points_expected = get_expected_shear_shape(
        length=length, width=wg_width, shear_angle=shear_angle
    )
    poly_actual = shear_waveguide_symmetric.get_polygons(by_spec=(1, 0))[0]
    assert_polygon_equals(points_expected, poly_actual)


def test_points_are_correct_long():
    wg_width = 28
    length = 100
    P = gf.path.straight(length=length)
    shear_waveguide_symmetric = gf.path.extrude(
        P,
        {"cross_section": "strip", "settings": {"width": wg_width}},
        shear_angle_start=DEMO_PORT_ANGLE,
        shear_angle_end=DEMO_PORT_ANGLE,
    )

    shear_angle = DEMO_PORT_ANGLE

    points_expected = get_expected_shear_shape(
        length=length, width=wg_width, shear_angle=shear_angle
    )
    poly_actual = shear_waveguide_symmetric.get_polygons(by_spec=(1, 0))[0]
    assert_polygon_equals(points_expected, poly_actual)


def test_points_are_correct_multi_layer():
    length = 1000
    P = gf.path.straight(length=length)

    s = gf.Section(width=30, offset=0, layer=gf.LAYER.SLAB90, name="slab")
    X1 = gf.CrossSection(
        width=1,
        offset=0,
        layer=gf.LAYER.WG,
        name="core",
        port_names=("o1", "o2"),
        sections=[s],
    )
    shear_waveguide_symmetric = gf.path.extrude(
        P, X1, shear_angle_start=DEMO_PORT_ANGLE, shear_angle_end=DEMO_PORT_ANGLE
    )
    shear_angle = DEMO_PORT_ANGLE

    for layer, wg_width in [(gf.LAYER.WG, 1), (gf.LAYER.SLAB90, 30)]:
        points_expected = get_expected_shear_shape(
            length=length, width=wg_width, shear_angle=shear_angle
        )
        poly_actual = shear_waveguide_symmetric.get_polygons(by_spec=layer)[0]
        assert_polygon_equals(points_expected, poly_actual)


if __name__ == "__main__":
    test_mate_on_shear_xor_empty_curve()
    # P = gf.path.euler()
    # curve = gf.path.extrude(P, "strip")

    # angle = 15
    # P = gf.path.euler()
    # curve_sheared1 = gf.path.extrude(P, "strip", shear_angle_end=angle)
    # curve_sheared2 = gf.path.extrude(P, "strip", shear_angle_start=angle)

    # two_straights = gf.Component()
    # c1 = two_straights << curve
    # c2 = two_straights << curve
    # c2.connect("o1", c1.ports["o2"])

    # two_shears = gf.Component()
    # c1 = two_shears << curve_sheared1
    # c2 = two_shears << curve_sheared2
    # c2.connect("o1", c1.ports["o2"])

    # xor = gf.geometry.xor_diff(two_straights, two_shears, precision=1e-2)
    # assert not xor.layers, f"{xor.layers}"

    # two_straights.show()
    # two_shears.show()

    P = gf.path.euler()
    curve = gf.path.extrude(P, "strip")

    angle = 15
    P = gf.path.euler()
    curve_sheared1 = gf.path.extrude(P, "strip", shear_angle_end=angle)
    curve_sheared2 = gf.path.extrude(P, "strip", shear_angle_start=angle)

    two_straights = gf.Component()
    c1 = two_straights << curve
    c2 = two_straights << curve
    c2.connect("o1", c1.ports["o2"])

    two_shears = gf.Component()
    c1 = two_shears << curve_sheared1
    c2 = two_shears << curve_sheared2
    c2.connect("o1", c1.ports["o2"])

    xor = gf.geometry.xor_diff(two_straights, two_shears, precision=1e-2)
    xor.show()



import uuid
import warnings
from typing import Callable, Dict, List, Optional, Tuple, Union

import gdspy
import numpy as np
from numpy import bool_, ndarray

import gdsfactory as gf
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.cross_section import strip
from gdsfactory.geometry.functions import angles_deg
from gdsfactory.port import Port, select_ports_list
from gdsfactory.routing.get_route_sbend import get_route_sbend
from gdsfactory.snap import snap_to_grid
from gdsfactory.tech import LAYER
from gdsfactory.types import (
    ComponentSpec,
    Coordinate,
    Coordinates,
    CrossSection,
    CrossSectionSpec,
    Layer,
    Layers,
    MultiCrossSectionAngleSpec,
    Route,
)

TOLERANCE = 0.001
DEG2RAD = np.pi / 180
RAD2DEG = 1 / DEG2RAD

O2D = {0: "East", 180: "West", 90: "North", 270: "South"}


class RouteWarning(UserWarning):
    pass


class RouteError(ValueError):
    pass


def sign(x: float) -> int:
    return -1 if x < 0 else 1


def _get_unique_port_facing(
    ports: Dict[str, Port],
    orientation: float = 0,
    layer: Union[Layer, Layers] = (1, 0),
) -> List[Port]:
    """Ensures there is only one port."""
    ports_selected = []
    if isinstance(layer, list):
        for _layer in layer:
            ports_selected = select_ports_list(
                ports=ports, orientation=orientation, layer=_layer
            )
            if ports_selected:
                break
    else:
        ports_selected = select_ports_list(
            ports=ports, orientation=orientation, layer=layer
        )

    if len(ports_selected) > 1:
        orientation %= 360
        direction = O2D[orientation]
        for port in ports_selected:
            print(port)
        raise ValueError(
            f"_get_unique_port_facing: \n\
            should have only one port facing {direction}\n\
            Got {len(ports_selected)} {[port.name for port in ports_selected]}"
        )

    return ports_selected


def _get_bend_ports(
    bend: Component,
    orientation: float = 0,
    layer: Union[Layer, Layers] = (1, 0),
) -> List[Port]:
    """Returns West and North facing ports for bend.

    Any standard bend/corner has two ports: one facing west and one
    facing north Returns these two ports in this order.

    """
    ports = bend.ports

    p_w = _get_unique_port_facing(ports=ports, orientation=180, layer=layer)
    p_n = _get_unique_port_facing(ports=ports, orientation=90, layer=layer)

    return p_w + p_n


def _get_straight_ports(
    straight: Component,
    layer: Tuple[int, int] = (1, 0),
) -> List[Port]:
    """Return West and east facing ports for straight waveguide.

    Any standard straight wire/straight has two ports: one facing west
    and one facing east

    """
    ports = straight.ports

    p_w = _get_unique_port_facing(ports=ports, orientation=180, layer=layer)
    p_e = _get_unique_port_facing(ports=ports, orientation=0, layer=layer)

    return p_w + p_e


def gen_sref(
    structure: Component,
    rotation_angle: float,
    x_reflection: bool,
    port_name: str,
    position: Coordinate,
) -> ComponentReference:
    """Place reference of `port_name` of `structure` at `position`.

    Keep this convention, otherwise phidl port transform won't work
    - 1 Mirror
    - 2 Rotate
    - 3 Move

    """
    position = np.array(position)

    if port_name is None:
        port_position = np.array([0, 0])
    else:
        port_position = structure.ports[port_name].center

    ref = gf.ComponentReference(component=structure, origin=(0, 0))

    if x_reflection:  # Vertical mirror: Reflection across x-axis
        y0 = port_position[1]
        ref.reflect(p1=(0, y0), p2=(1, y0))

    ref.rotate(rotation_angle, center=port_position)
    ref.move(port_position, position)
    return ref


def _is_vertical(p0: ndarray, p1: ndarray) -> bool_:
    return np.abs(p0[0] - p1[0]) < TOLERANCE


def _is_horizontal(p0: ndarray, p1: ndarray) -> bool_:
    return np.abs(p0[1] - p1[1]) < TOLERANCE


def get_straight_distance(p0: ndarray, p1: ndarray) -> float:
    if _is_vertical(p0, p1):
        return np.abs(p0[1] - p1[1])
    if _is_horizontal(p0, p1):
        return np.abs(p0[0] - p1[0])

    raise RouteError(f"Waveguide points {p0} {p1} are not manhattan")


def transform(
    points: ndarray,
    translation: ndarray,
    angle_deg: int = 0,
    x_reflection: bool = False,
) -> ndarray:
    """Transform points.

    Args:
        points (np.array of shape (N,2) ): points to be transformed.
        translation (2d like array): translation vector.
        angle_deg: rotation angle.
        x_reflection (bool): if True, mirror the shape across the x axis  (y -> -y).

    """
    # Copy
    pts = points[:, :]

    # Translate
    pts = pts + np.array(translation)

    # Rotate
    c = np.cos(DEG2RAD * angle_deg)
    s = np.sin(DEG2RAD * angle_deg)
    rotation_matrix = np.array([[c, s], [-s, c]])
    pts = np.dot(pts, rotation_matrix)

    # Mirror
    if x_reflection:
        pts[:, 1] = -pts[:, 1]

    return pts


def reverse_transform(
    points: ndarray,
    translation: Coordinate = (0, 0),
    angle_deg: int = 0,
    x_reflection: bool = False,
) -> ndarray:
    """Args are the following.

    points (np.array of shape (N,2) ): points to be transformed.
    translation (2d like array): translation vector.
    angle_deg: rotation angle.
    x_reflection: if True, mirror the shape across the x axis  (y -> -y).
    """
    angle_deg = -angle_deg

    # Copy
    pts = points[:, :]

    # Mirror
    if x_reflection:
        pts[:, 1] = -pts[:, 1]

    # Rotate
    c = np.cos(DEG2RAD * angle_deg)
    s = np.sin(DEG2RAD * angle_deg)
    rotation_matrix = np.array([[c, s], [-s, c]])
    pts = np.dot(pts, rotation_matrix)

    # Translate
    pts = pts - np.array(translation)
    return pts


def _generate_route_manhattan_points(
    input_port: Port,
    output_port: Port,
    bs1: float,
    bs2: float,
    start_straight_length: float = 0.01,
    end_straight_length: float = 0.01,
    min_straight_length: float = 0.01,
) -> ndarray:
    """Return list of ports for the route.

    Args:
        input_port:
        output_port:
        bs1: bend size
        bs2: bend size
        start_straight_length: in um.
        end_straight_length: in um.
        min_straight_length: in um.

    """
    threshold = TOLERANCE

    # transform I/O to the case where output is at (0, 0) pointing east (180)
    p_input = input_port.center
    p_output = np.array(output_port.center)
    pts_io = np.stack([p_input, p_output], axis=0)
    angle = output_port.orientation

    if output_port.orientation is None and input_port.orientation is None:
        x0, y0 = p_input
        x2, y2 = p_output
        p1 = (x0, y2)
        points = np.array([p_input, p1, p_output])
    elif input_port.orientation is None:
        raise ValueError("input_port orientation is None")

    elif output_port.orientation is None:
        raise ValueError("output_port orientation is None")

    else:
        bend_orientation = -angle + 180
        transform_params = dict(
            translation=-p_output, angle_deg=bend_orientation, x_reflection=False
        )

        _pts_io = transform(pts_io, **transform_params)
        p = _pts_io[0, :]
        _p_output = _pts_io[1, :]

        a = int(input_port.orientation + bend_orientation) % 360
        s = start_straight_length
        count = 0
        points = [p]

        while True:
            count += 1
            if count > 40:
                raise AttributeError(
                    f"Too many iterations for in {input_port} -> out {output_port}"
                )
            # not ready for final bend: go over possibilities
            sigp = np.sign(p[1])
            if not sigp:
                sigp = 1
            if a % 360 == 0:
                # same directions
                if abs(p[1]) < threshold and p[0] <= threshold:
                    # Reach the output!
                    points += [_p_output]
                    break
                elif (
                    p[0] + (bs1 + bs2 + end_straight_length + s) < threshold
                    and abs(p[1]) - (bs1 + bs2 + min_straight_length) > -threshold
                ):
                    # sufficient space for S-bend
                    p = (-end_straight_length - bs2, p[1])
                    a = -sigp * 90
                elif (
                    p[0]
                    + (
                        2 * bs1
                        + 2 * bs2
                        + end_straight_length
                        + s
                        + min_straight_length
                    )
                    < threshold
                ):
                    # sufficient distance to move aside
                    p = (p[0] + s + bs1, p[1])
                    a = -sigp * 90
                elif (
                    abs(p[1]) - (2 * bs1 + 2 * bs2 + 2 * min_straight_length)
                    > -threshold
                ):
                    p = (p[0] + s + bs1, p[1])
                    a = -sigp * 90
                else:
                    p = (p[0] + s + bs1, p[1])
                    a = sigp * 90

            elif a == 180:
                # opposite directions
                if abs(p[1]) - (bs1 + bs2 + min_straight_length) > -threshold:
                    # far enough: U-turn
                    p = (min(p[0] - s, -end_straight_length) - bs2, p[1])
                else:
                    # more complex turn
                    p = (
                        min(
                            p[0] - s - bs1,
                            -end_straight_length - min_straight_length - 2 * bs1 - bs2,
                        ),
                        p[1],
                    )
                a = -sigp * 90
            elif a % 180 == 90:
                siga = -np.sign((a % 360) - 180)
                if not siga:
                    siga = 1

                if ((-p[1] * siga) - (s + bs2) > -threshold) and (
                    -p[0] - (end_straight_length + bs2)
                ) > -threshold:
                    # simple case: one right angle to the end
                    p = (p[0], 0)
                    a = 0
                elif (p[1] * siga) <= threshold and p[0] + (
                    end_straight_length + bs1
                ) > -threshold:
                    # go to the west, and then turn upward
                    # this will sometimes result in too sharp bends, but there is no
                    # avoiding this!

                    _y = min(
                        max(
                            min(min_straight_length, 0.5 * abs(p[1])),
                            abs(p[1]) - s - bs1,
                        ),
                        bs1 + bs2 + min_straight_length,
                    )

                    p = (p[0], sigp * _y)
                    if count == 1:  # take care of the start_straight case
                        p = (p[0], -sigp * max(start_straight_length, _y))

                    a = 180
                elif (
                    -p[0] - (end_straight_length + 2 * bs1 + bs2 + min_straight_length)
                    > -threshold
                ):
                    # go sufficiently up, and then east
                    p = (
                        p[0],
                        siga
                        * max(p[1] * siga + s + bs1, bs1 + bs2 + min_straight_length),
                    )
                    a = 0

                elif -p[0] - (end_straight_length + bs2) > -threshold:
                    # make vertical S-bend to get sufficient room for movement
                    points += [(p[0], p[1] + siga * (bs2 + s))]
                    p = (
                        min(
                            p[0] - bs1 + bs2 + min_straight_length,
                            -2 * bs1 - bs2 - end_straight_length - min_straight_length,
                        ),
                        p[1] + siga * (bs2 + s),
                    )
                    # `a` remains the same
                else:
                    # no viable solution for this case. May result in crossed straights
                    p = (p[0], p[1] + sigp * (s + bs1))
                    a = 180
            points += [p]
            s = min_straight_length + bs1
        points = np.stack([np.array(_p) for _p in points], axis=0)
        points = reverse_transform(points, **transform_params)
    return points


def _get_bend_reference_parameters(
    p0: ndarray,
    p1: ndarray,
    p2: ndarray,
    bend_cell: Component,
    port_layer: Union[Layer, List[Layer]],
) -> Tuple[ndarray, int, bool]:
    """Returns bend reference settings.

    Args:
        p0: starting port waypoints.
        p1: middle port waypoints.
        p2: end port points.
        bend_cell: bend component.
        port_layer: for the port.

    8 possible configurations
    First mirror, Then rotate

    Returns:
    .. code::

       p2        p2
        |         |
        | C     A |
       p1-<-p0->-p1      dp1 horizontal
        |         |      dp2 vertical
        | D     B |
       p2        p2


       p2-<-p1->-p2
            |
          F |  E         dp1 vertical
            p0           dp2 horizontal
          H |  G
            |
       p2-<-p1->-p2

    """
    # is_horizontal(dp1), s1, s2 : transform (rotation, vertical mirror)
    transforms_map = {
        (True, 1, 1): (0, False),  # A No transform
        (True, 1, -1): (0, True),  # B Vertical mirror
        (True, -1, 1): (180, True),  # C Horizontal mirror
        (True, -1, -1): (180, False),  # D R180
        (False, 1, 1): (90, True),  # E R90 + vertical mirror
        (False, 1, -1): (90, False),  # F R270
        (False, -1, 1): (270, False),  # G R270
        (False, -1, -1): (270, True),  # H R270 + vertical mirror
    }

    b1, b2 = (p.center for p in _get_bend_ports(bend=bend_cell, layer=port_layer))

    bsx = b2[0] - b1[0]
    bsy = b2[1] - b1[1]

    dp1 = p1 - p0
    dp2 = p2 - p1
    is_h_dp1 = np.abs(dp1[1]) < TOLERANCE

    if is_h_dp1:
        xd1 = dp1[0]
        yd2 = dp2[1]
        s1 = sign(xd1)
        s2 = sign(yd2)

        bend_origin = p1 - np.array([s1 * bsx, 0])

    else:
        yd1 = dp1[1]
        xd2 = dp2[0]
        s1 = int(sign(yd1))
        s2 = int(sign(xd2))

        bend_origin = p1 - np.array([0, s1 * bsy])

    t = transforms_map[(is_h_dp1, s1, s2)]

    return bend_origin, t[0], t[1]


def make_ref(component_factory: Callable) -> Callable:
    def _make_ref(*args, **kwargs):
        return component_factory(*args, **kwargs).ref()

    return _make_ref


def remove_flat_angles(points: ndarray) -> ndarray:
    a = angles_deg(np.vstack(points))
    da = a - np.roll(a, 1)
    da = np.mod(np.round(da, 3), 180)

    # To make sure we do not remove points at the edges
    da[0] = 1
    da[-1] = 1

    to_rm = list(np.where(np.abs(da[:-1]) < 1e-9)[0])
    if isinstance(points, list):
        while to_rm:
            i = to_rm.pop()
            points.pop(i)

    else:
        points = points[da != 0]

    return points


def get_route_error(
    points,
    cross_section: Optional[CrossSection] = None,
    layer_path: Layer = LAYER.ERROR_PATH,
    layer_label: Layer = LAYER.TEXT,
    layer_marker: Layer = LAYER.ERROR_MARKER,
    references: Optional[List[ComponentReference]] = None,
    with_sbend: bool = False,
) -> Route:
    """Returns route with error markers.

    Args:
        points: route waypoints.
        cross_section: Optional cross_section.
        layer_path: for the error.
        layer_label: for the labels.
        layer_marker: for point markers.
        references: optional list of references.
        with_sbend: if True raises Error so we can use it in try, except
            if False raises Warning.
    """
    layer_path = gf.get_layer(layer_path)
    layer_label = gf.get_layer(layer_label)
    layer_marker = gf.get_layer(layer_marker)

    width = cross_section.width if cross_section else 10

    if with_sbend:
        raise RouteError(f"route error for points {points}")
    else:
        warnings.warn(f"Route error for points {points}", RouteWarning)

    c = Component(f"route_{uuid.uuid4()}"[:16])
    path = gdspy.FlexPath(
        points,
        width=width,
        gdsii_path=True,
        layer=layer_path[0],
        datatype=layer_path[1],
    )
    c.add(path)
    ref = ComponentReference(c)
    port1 = gf.Port(
        name="p1", center=points[0], width=width, layer=layer_path, orientation=0
    )
    port2 = gf.Port(
        name="p2", center=points[1], width=width, layer=layer_path, orientation=0
    )

    point_marker = gf.components.rectangle(
        size=(width * 2, width * 2), centered=True, layer=layer_marker
    )
    point_markers = [point_marker.ref(position=point) for point in points] + [ref]
    labels = [
        gf.Label(
            text=str(i), position=point, layer=layer_label[0], texttype=layer_label[1]
        )
        for i, point in enumerate(points)
    ]

    references = references or []
    references += point_markers
    return Route(references=references, ports=[port1, port2], length=-1, labels=labels)


def round_corners(
    points: Coordinates,
    straight: ComponentSpec = straight_function,
    bend: ComponentSpec = bend_euler,
    taper: Optional[ComponentSpec] = None,
    straight_fall_back_no_taper: Optional[ComponentSpec] = None,
    mirror_straight: bool = False,
    straight_ports: Optional[List[str]] = None,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = strip,
    on_route_error: Callable = get_route_error,
    with_point_markers: bool = False,
    snap_to_grid_nm: Optional[int] = 1,
    with_sbend: bool = False,
    **kwargs,
) -> Route:
    """Returns Route.

    - reference list with rounded straight route from a list of manhattan points.
    - ports: Tuple of ports.
    - length: route length (float).

    Args:
        points: manhattan route defined by waypoints.
        bend90: the bend to use for 90Deg turns.
        straight: the straight library to use to generate straight portions.
        taper: taper for straight portions. If None, no tapering.
        straight_fall_back_no_taper: in case there is no space for two tapers.
        mirror_straight: mirror_straight waveguide.
        straight_ports: port names for straights. If None finds them automatically.
        cross_section: spec.
        on_route_error: function to run when route fails.
        with_point_markers: add route points markers (easy for debugging).
        snap_to_grid_nm: nm to snap to grid.
        with_sbend: add sbend in case there are routing errors.
        kwargs: cross_section settings.

    """
    from gdsfactory.pdk import get_layer

    multi_cross_section = isinstance(cross_section, list)
    if multi_cross_section:
        x = [gf.get_cross_section(xsection[0], **kwargs) for xsection in cross_section]
        layer = [_x.layer for _x in x]
    else:
        x = gf.get_cross_section(cross_section, **kwargs)
        layer = x.layer

    layer = get_layer(layer)
    points = (
        gf.snap.snap_to_grid(points, nm=snap_to_grid_nm) if snap_to_grid_nm else points
    )

    references = []

    bend90 = (
        bend
        if isinstance(bend, Component)
        else gf.get_component(bend, cross_section=cross_section, **kwargs)
    )

    # bsx = bsy = _get_bend_size(bend90)
    auto_widen = (
        x.auto_widen if not isinstance(x, list) else [_x.auto_widen for _x in x]
    )
    auto_widen_minimum_length = (
        x.auto_widen_minimum_length
        if not isinstance(x, list)
        else [_x.auto_widen_minimum_length for _x in x]
    )
    taper_length = (
        x.taper_length if not isinstance(x, list) else [_x.taper_length for _x in x]
    )
    width = x.width if not isinstance(x, list) else [_x.width for _x in x]
    width_wide = (
        x.width_wide if not isinstance(x, list) else [_x.width_wide for _x in x]
    )

    if isinstance(cross_section, list):
        taper = None
    elif taper is None:
        taper = taper_function(
            cross_section=cross_section,
            width1=width,
            width2=width_wide,
            length=taper_length,
        )
    elif not isinstance(taper, Component):
        taper = gf.get_component(taper, cross_section=cross_section, **kwargs)

    # If there is a taper, make sure its length is known
    if taper and isinstance(taper, Component):
        if "length" not in taper.info:
            _taper_ports = list(taper.ports.values())
            taper.info["length"] = _taper_ports[-1].x - _taper_ports[0].x

    straight_fall_back_no_taper = straight_fall_back_no_taper or straight

    # Remove any flat angle, otherwise the algorithm won't work
    points = remove_flat_angles(points)
    points = np.array(points)

    straight_sections = []  # (p0, angle, length)
    p0_straight = points[0]
    p1 = points[1]

    total_length = 0  # Keep track of the total path length

    if not bend90.info.get("length"):
        raise ValueError(f"bend {bend90} needs to have bend.info['length'] defined")

    bend_length = bend90.info["length"]

    dp = p1 - p0_straight
    bend_orientation = None
    if _is_vertical(p0_straight, p1):
        if dp[1] > 0:
            bend_orientation = 90
        elif dp[1] < 0:
            bend_orientation = 270
    elif _is_horizontal(p0_straight, p1):
        if dp[0] > 0:
            bend_orientation = 0
        elif dp[0] < 0:
            bend_orientation = 180

    if bend_orientation is None:
        print(f"bend_orientation is None {p0_straight} {p1}")
        return on_route_error(
            points=points,
            cross_section=x if not multi_cross_section else None,
            with_sbend=with_sbend,
        )

    try:
        pname_west, pname_north = (
            p.name for p in _get_bend_ports(bend=bend90, layer=layer)
        )
    except ValueError as exc:
        raise ValueError(
            f"Did not find 2 ports on layer {layer}. Got {list(bend90.ports.values())}"
        ) from exc
    n_o_bends = points.shape[0] - 2
    total_length += n_o_bends * bend_length

    previous_port_point = points[0]
    bend_points = [previous_port_point]

    # Add bend sections and record straight-section information
    for i in range(1, points.shape[0] - 1):
        bend_origin, rotation, x_reflection = _get_bend_reference_parameters(
            points[i - 1], points[i], points[i + 1], bend90, layer
        )
        bend_ref = gen_sref(bend90, rotation, x_reflection, pname_west, bend_origin)
        references.append(bend_ref)

        dx_points = points[i][0] - points[i - 1][0]
        dy_points = points[i][1] - points[i - 1][1]

        if abs(dx_points) < TOLERANCE:
            matching_ports = [
                port
                for port in bend_ref.ports.values()
                if np.isclose(port.x, points[i][0])
            ]

        if abs(dy_points) < TOLERANCE:
            matching_ports = [
                port
                for port in bend_ref.ports.values()
                if np.isclose(port.y, points[i][1])
            ]

        if matching_ports:
            next_port = matching_ports[0]
            other_port_name = set(bend_ref.ports.keys()) - {next_port.name}
            other_port = bend_ref.ports[list(other_port_name)[0]]
            bend_points.append(next_port.center)
            bend_points.append(other_port.center)
            previous_port_point = other_port.center

        try:
            straight_sections += [
                (
                    p0_straight,
                    bend_orientation,
                    get_straight_distance(p0_straight, bend_origin),
                )
            ]
        except RouteError as e:
            print(e)
            on_route_error(
                points=(p0_straight, bend_origin),
                cross_section=x if not multi_cross_section else None,
                references=references,
                with_sbend=with_sbend,
            )

        p0_straight = bend_ref.ports[pname_north].center
        bend_orientation = bend_ref.ports[pname_north].orientation

    bend_points.append(points[-1])

    try:
        straight_sections += [
            (
                p0_straight,
                bend_orientation,
                get_straight_distance(p0_straight, points[-1]),
            )
        ]
    except RouteError as e:
        print(e)
        on_route_error(
            points=(p0_straight, points[-1]),
            cross_section=x if not multi_cross_section else None,
            references=references,
            with_sbend=with_sbend,
        )

    # ensure bend connectivity
    for i, point in enumerate(points[:-1]):
        sx = np.sign(points[i + 1][0] - point[0])
        sy = np.sign(points[i + 1][1] - point[1])
        bsx = np.sign(bend_points[2 * i + 1][0] - bend_points[2 * i][0])
        bsy = np.sign(bend_points[2 * i + 1][1] - bend_points[2 * i][1])
        if bsx * sx == -1 or bsy * sy == -1:
            return on_route_error(
                points=points,
                cross_section=x if not multi_cross_section else None,
                references=references,
                with_sbend=with_sbend,
            )

    wg_refs = []
    for straight_origin, angle, length in straight_sections:
        if isinstance(cross_section, list):
            for section, angles in cross_section:
                if angle in angles:
                    xsection = section
                    break
        else:
            xsection = cross_section
        x = gf.get_cross_section(xsection, **kwargs)

        with_taper = False
        # wg_width = list(bend90.ports.values())[0].width
        length = snap_to_grid(length)
        total_length += length

        if isinstance(cross_section, list):
            wg = gf.get_component(
                straight_fall_back_no_taper,
                length=length,
                cross_section=xsection,
                **kwargs,
            )
        elif auto_widen and length > auto_widen_minimum_length and width_wide:
            # Taper starts where straight would have started
            with_taper = True
            length = length - 2 * taper_length
            taper_origin = straight_origin

            pname_west, pname_east = (
                p.name for p in _get_straight_ports(taper, layer=layer)
            )
            taper_ref = taper.ref(
                position=taper_origin, port_id=pname_west, rotation=angle
            )

            references.append(taper_ref)
            wg_refs += [taper_ref]

            # Update start straight position
            straight_origin = taper_ref.ports[pname_east].center

            # Straight waveguide
            kwargs_wide = kwargs.copy()
            kwargs_wide.update(width=width_wide)

            if callable(cross_section):
                cross_section_wide = gf.partial(cross_section, **kwargs_wide)
            else:
                cross_section_wide = x.copy(width=width_wide)
            wg = gf.get_component(
                straight, length=length, cross_section=cross_section_wide
            )
        else:
            wg = gf.get_component(
                straight_fall_back_no_taper,
                length=length,
                cross_section=xsection,
                **kwargs,
            )

        if straight_ports is None:
            straight_ports = [p.name for p in _get_straight_ports(wg, layer=layer)]
        pname_west, pname_east = straight_ports

        wg_ref = wg.ref()
        wg_ref.move(wg.ports[pname_west], (0, 0))
        if mirror_straight:
            wg_ref.reflect_v(list(wg_ref.ports.values())[0].name)

        wg_ref.rotate(angle)
        wg_ref.move(straight_origin)

        if length > 0:
            references.append(wg_ref)
            wg_refs += [wg_ref]

        port_index_out = 1
        if with_taper:
            # Second taper:
            # Origin at end of straight waveguide, starting from east side of taper

            taper_origin = wg_ref.ports[pname_east]
            pname_west, pname_east = (
                p.name for p in _get_straight_ports(taper, layer=layer)
            )

            taper_ref = taper.ref(
                position=taper_origin,
                port_id=pname_east,
                rotation=angle + 180,
                v_mirror=True,
            )
            # references += [
            #     gf.Label(
            #         text=f"a{angle}",
            #         position=taper_ref.center,
            #         layer=2,
            #         texttype=0,
            #     )
            # ]
            references.append(taper_ref)
            wg_refs += [taper_ref]
            port_index_out = 0

    if with_point_markers:
        route = get_route_error(
            points, cross_section=x if not multi_cross_section else None
        )
        references += route.references

    port_input = list(wg_refs[0].ports.values())[0]
    port_output = list(wg_refs[-1].ports.values())[port_index_out]
    length = snap_to_grid(float(total_length))
    return Route(references=references, ports=(port_input, port_output), length=length)


def generate_manhattan_waypoints(
    input_port: Port,
    output_port: Port,
    start_straight_length: Optional[float] = None,
    end_straight_length: Optional[float] = None,
    min_straight_length: Optional[float] = None,
    bend: ComponentSpec = bend_euler,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = strip,
    **kwargs,
) -> ndarray:
    """Return waypoints for a Manhattan route between two ports.

    Args:
        input_port: source port.
        output_port: destination port.
        start_straight_length: Optional start length.
        end_straight_length: in um.
        min_straight_length: in um.
        bend: bend spec.
        cross_section: spec.
        kwargs: cross_section settings.

    """
    if "straight" in kwargs.keys():
        _ = kwargs.pop("straight")

    bend90 = (
        bend
        if isinstance(bend, Component)
        else gf.get_component(bend, cross_section=cross_section, **kwargs)
    )
    if isinstance(cross_section, list):
        x = [gf.get_cross_section(xsection[0], **kwargs) for xsection in cross_section]
        start_straight_length = start_straight_length or min(_x.min_length for _x in x)
        end_straight_length = end_straight_length or min(_x.min_length for _x in x)
        min_straight_length = min_straight_length or min(_x.min_length for _x in x)
    else:
        x = gf.get_cross_section(cross_section, **kwargs)
        start_straight_length = start_straight_length or x.min_length
        end_straight_length = end_straight_length or x.min_length
        min_straight_length = min_straight_length or x.min_length
    bsx = bsy = _get_bend_size(bend90)
    points = _generate_route_manhattan_points(
        input_port,
        output_port,
        bsx,
        bsy,
        start_straight_length,
        end_straight_length,
        min_straight_length,
    )
    return points


def _get_bend_size(bend90: Component):
    p1, p2 = list(bend90.ports.values())[:2]
    bsx = abs(p2.x - p1.x)
    bsy = abs(p2.y - p1.y)
    return max(bsx, bsy)


def route_manhattan(
    input_port: Port,
    output_port: Port,
    straight: ComponentSpec = straight_function,
    taper: Optional[ComponentSpec] = None,
    start_straight_length: Optional[float] = None,
    end_straight_length: Optional[float] = None,
    min_straight_length: Optional[float] = None,
    bend: ComponentSpec = bend_euler,
    with_sbend: bool = True,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = strip,
    with_point_markers: bool = False,
    on_route_error: Callable = get_route_error,
    **kwargs,
) -> Route:
    """Generates the Manhattan waypoints for a route.

    Then creates the straight, taper and bend references that define the
    route, or create an SBend route.

    Args:
        input_port: input.
        output_port: output.
        straight: function.
        taper: add taper.
        start_straight_length: in um.
        end_straight_length: in um.
        min_straight_length: min length of straight for any intermediate segment.
        bend: bend spec.
        with_sbend: add sbend in case there are routing errors.
        cross_section: spec.
        with_point_markers: add point markers in the route.
        kwargs: cross_section settings.

    """
    if isinstance(cross_section, list):
        x = [gf.get_cross_section(xsection[0], **kwargs) for xsection in cross_section]
        start_straight_length = start_straight_length or min(_x.min_length for _x in x)
        end_straight_length = end_straight_length or min(_x.min_length for _x in x)
        min_straight_length = min_straight_length or min(_x.min_length for _x in x)
    else:
        x = gf.get_cross_section(cross_section, **kwargs)
        start_straight_length = start_straight_length or x.min_length
        end_straight_length = end_straight_length or x.min_length
        min_straight_length = min_straight_length or x.min_length

    try:
        points = generate_manhattan_waypoints(
            input_port,
            output_port,
            start_straight_length=start_straight_length,
            end_straight_length=end_straight_length,
            min_straight_length=min_straight_length,
            bend=bend,
            cross_section=x,
        )
        route = round_corners(
            points=points,
            straight=straight,
            taper=taper,
            bend=bend,
            cross_section=x,
            with_point_markers=with_point_markers,
            with_sbend=with_sbend,
        )
        return route

    except RouteError:
        if with_sbend:
            return get_route_sbend(input_port, output_port, cross_section=x)

    return get_route_error(points=points, with_sbend=False)


if __name__ == "__main__":
    c = gf.Component("pads_route_from_steps")
    pt = c << gf.components.pad_array(orientation=270, columns=3)
    pb = c << gf.components.pad_array(orientation=90, columns=3)
    pt.move((100, 200))
    route = gf.routing.get_route_from_steps(
        pt.ports["e11"],
        pb.ports["e11"],
        steps=[
            {"y": 100},
        ],
        cross_section=gf.cross_section.metal3,
        bend=gf.components.wire_corner,
    )
    c.add(route.references)
    c.show(show_ports=True)



import pathlib
import re

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from gdsfactory.types import PathType


def pandas_to_numpy(df: pd.DataFrame, port_map=None) -> np.ndarray:
    """Converts a pandas CSV sparameters into a numpy array.

    Fundamental mode starts at 0.
    """
    s_headers = sorted({c[:-1] for c in df.columns if c.lower().startswith("s")})
    idxs = sorted({idx for c in s_headers for idx in _s_header_to_port_idxs(c)})

    if port_map is None:
        port_map = {f"o{i}@0": i for i in idxs}
    rev_port_map = {i: p for p, i in port_map.items()}
    assert len(rev_port_map) == len(
        port_map
    ), "Duplicate port indices found in port_map"

    s_map = {
        s: tuple(rev_port_map[i] for i in _s_header_to_port_idxs(s)) for s in s_headers
    }

    dfs = {
        s: df[["wavelengths", f"{s}m", f"{s}a"]]
        .copy()
        .rename(columns={f"{s}m": "magnitude", f"{s}a": "phase"})
        for s in s_map
    }

    S = dict(wavelengths=df["wavelengths"].values)
    for key, df in dfs.items():
        pm1, pm2 = s_map[key]
        (p1, m1), (p2, m2) = pm1.split("@"), pm2.split("@")
        name = f"{p1}@{m1},{p2}@{m2}"
        S[name] = df["magnitude"].values * np.exp(1j * df["phase"].values)

    return S


def csv_to_npz(filepath: PathType) -> pathlib.Path:
    df = pd.read_csv(filepath)
    sp = pandas_to_numpy(df)
    filepath_npz = pathlib.Path(filepath).with_suffix(".npz")
    np.savez_compressed(filepath_npz, **sp)
    return filepath_npz


def convert_directory_csv_to_npz(dirpath: PathType) -> None:
    dirpath = pathlib.Path(dirpath)
    for filepath in tqdm(dirpath.glob("**/*.csv")):
        try:
            csv_to_npz(filepath)
        except Exception as e:
            print(filepath)
            print(e)


def _s_header_to_port_idxs(s):
    s = re.sub("[^0-9]", "", s)
    inp, out = (int(i) for i in s)
    return inp, out


if __name__ == "__main__":
    # import matplotlib.pyplot as plt

    import gdsfactory as gf

    filepath = gf.CONFIG["sparameters"] / "mmi1x2_d542be8a.csv"
    filepath = gf.CONFIG["sparameters"] / "mmi1x2_00cc8908.csv"
    filepath = gf.CONFIG["sparameters"] / "mmi1x2_1f90b7ca.csv"  # lumerical

    convert_directory_csv_to_npz(gf.CONFIG["sparameters"])

    # s = csv_to_npz(filepath)
    # plt.plot(s["wavelengths"], np.abs(s["o1@0,o2@0"]) ** 2)
    # plt.plot(s["wavelengths"], np.abs(s["o1@0,o3@0"]) ** 2)
    # plt.show()



from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
from numpy import float64, ndarray

from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.geometry.functions import remove_identicals
from gdsfactory.port import Port
from gdsfactory.routing.get_route import get_route_from_waypoints
from gdsfactory.routing.manhattan import (
    generate_manhattan_waypoints,
    remove_flat_angles,
)
from gdsfactory.routing.path_length_matching import path_length_matched_points
from gdsfactory.routing.route_ports_to_side import route_ports_to_side
from gdsfactory.types import ComponentSpec, Route


def _groups(
    ports: List[Port], cut: float64, axis: str = "X"
) -> Union[Tuple[List[Port], List[Any]], Tuple[List[Port], List[Port]]]:
    if axis == "Y":
        group1 = [p for p in ports if p.x <= cut]
        group2 = [p for p in ports if p.x > cut]
    else:
        group1 = [p for p in ports if p.y <= cut]
        group2 = [p for p in ports if p.y > cut]
    return group1, group2


def get_bundle_udirect(
    ports1: List[Port],
    ports2: List[Port],
    route_filter: Callable = get_route_from_waypoints,
    separation: float = 5.0,
    start_straight_length: float = 0.01,
    end_straight_length: float = 0.01,
    bend: ComponentSpec = bend_euler,
    path_length_match_loops: Optional[int] = None,
    path_length_match_extra_length: float = 0.0,
    path_length_match_modify_segment_i: int = -2,
    **kwargs,
) -> List[Route]:
    r"""Returns list of routes.

    Args:
        ports1: list of start ports.
        ports2: list of end ports.
        route_filter: filter to apply to the manhattan waypoints
            e.g `get_route_from_waypoints` for deep etch strip straight.
        separation: between straights.
        start_straight_length: in um.
        end_straight_length: in um.
        bend: bend spec.
        path_length_match_loops: Integer number of loops to add to bundle
            for path length matching (won't try to match if None).
        path_length_match_extra_length: Extra length to add
            to path length matching loops (requires path_length_match_loops != None).
        path_length_match_modify_segment_i: Index of straight segment to add path
            length matching loops to (requires path_length_match_loops != None).

    Returns:
        [route_filter(r) for r in routes] where routes is a list of lists of coordinates
        e.g with default `get_route_from_waypoints`,
        returns list of elements which can be added to a component


    Used for routing multiple ports back to a bundled input in a component

    ::

        X: start ports
        D: End ports

        On this example below, the axis is along X

                           X------\
                                  |
                           X----\ |
                          ...   | |
                           X--\ | |
                              | | |
          D-------------------/ | |
          ...                   | |
          D---------------------/ |
          D-----------------------
          D-----------------------\
          D---------------------\ |
          ...                   | |
          D-------------------\ | |
                              | | |
                           X--/ | |
                          ...   | |
                           X----/ |
                                  |
                           X------/
    """
    if "straight" in kwargs.keys():
        _ = kwargs.pop("straight")
    routes = _get_bundle_udirect_waypoints(
        ports1,
        ports2,
        separation=separation,
        start_straight_length=start_straight_length,
        end_straight_offset=end_straight_length,
        routing_func=generate_manhattan_waypoints,
        bend=bend,
        **kwargs,
    )
    if path_length_match_loops:
        routes = [np.array(route) for route in routes]
        routes = path_length_matched_points(
            routes,
            extra_length=path_length_match_extra_length,
            bend=bend,
            nb_loops=path_length_match_loops,
            modify_segment_i=path_length_match_modify_segment_i,
            # cross_section=cross_section,
            **kwargs,
        )

    return [route_filter(route, bend=bend, **kwargs) for route in routes]


def _get_bundle_udirect_waypoints(
    ports1: List[Port],
    ports2: List[Port],
    routing_func: Callable = generate_manhattan_waypoints,
    separation: float = 5.0,
    start_straight_length: float = 0.01,
    end_straight_length: float = 0.01,
    end_straight_offset: float = 0.0,
    start_straight_offset: float = 0.0,
    bend: ComponentSpec = bend_euler,
    **routing_func_params,
) -> List[ndarray]:

    nb_ports = len(ports1)
    for p in ports1:
        p.orientation = (
            p.orientation % 360 if p.orientation is not None else p.orientation
        )

    for p in ports2:
        p.orientation = (
            p.orientation % 360 if p.orientation is not None else p.orientation
        )

    if len(ports2) != nb_ports:
        raise ValueError(
            "Number of start ports should match number of end ports."
            f"Got {len(ports1)} {len(ports2)}"
        )
    if len({p.orientation for p in ports1 + ports2}) > 1:
        orientations1 = [p.orientation for p in ports1]
        orientations2 = [p.orientation for p in ports2]
        raise ValueError(
            "All ports should have the same orientation. "
            f"Got \n{orientations1}\n{orientations2}"
        )

    xs_end = [p.x for p in ports2]
    ys_end = [p.y for p in ports2]

    x_cut = 0.5 * (min(xs_end) + max(xs_end))
    y_cut = 0.5 * (min(ys_end) + max(ys_end))

    # Find axis
    angle_start = ports1[0].orientation

    if angle_start in [0, 180]:
        axis = "X"
        cut = y_cut
    else:
        axis = "Y"
        cut = x_cut

    # Get groups (below, above) or (left, right)
    group1, group2 = _groups(ports1, cut, axis=axis)

    # Sort ports to make them easy to connect
    if axis == "X":
        group1.sort(key=lambda p: -p.y)
        group2.sort(key=lambda p: p.y)

        ports2.sort(key=lambda p: p.y)
        end_group1 = ports2[: len(group1)]
        end_group2 = ports2[len(group1) :]
        end_group2.sort(key=lambda p: -p.y)

        xs_start = [p.x for p in ports1]

        if angle_start == 0:
            dx = xs_start[0] - xs_end[0]
        elif angle_start == 180:
            dx = xs_end[0] - xs_start[0]
        end_straight_length = max(end_straight_length, dx)

    elif axis == "Y":
        group1.sort(key=lambda p: -p.x)
        group2.sort(key=lambda p: p.x)

        ports2.sort(key=lambda p: p.x)
        end_group1 = ports2[: len(group1)]
        end_group2 = ports2[len(group1) :]
        end_group2.sort(key=lambda p: -p.x)

        ys_start = [p.y for p in ports1]

        if angle_start == 90:
            dy = ys_start[0] - ys_end[0]
        elif angle_start == 270 or angle_start is None:
            dy = ys_end[0] - ys_start[0]
        end_straight_length = max(end_straight_length, dy)

    # add offsets
    start_straight_length += start_straight_offset
    end_straight_length += end_straight_offset

    connections = []
    straight_len_end = end_straight_length
    straight_len_start = start_straight_length
    for p_start, p_end in zip(group1, end_group1):
        _c = routing_func(
            p_start,
            p_end,
            start_straight_length=straight_len_start,
            end_straight_length=straight_len_end,
            bend=bend,
            **routing_func_params,
        )
        connections += [_c]
        straight_len_end += separation
        straight_len_start += separation

    straight_len_end = end_straight_length
    straight_len_start = start_straight_length
    for p_start, p_end in zip(group2, end_group2):
        _c = routing_func(
            p_start,
            p_end,
            start_straight_length=straight_len_start,
            end_straight_length=straight_len_end,
            bend=bend,
            **routing_func_params,
        )
        connections += [_c]
        straight_len_end += separation
        straight_len_start += separation

    return connections


def get_bundle_uindirect(
    ports1: List[Port],
    ports2: List[Port],
    route_filter: Callable = get_route_from_waypoints,
    separation: float = 5.0,
    extension_length: float = 0.0,
    start_straight_length: float = 0.01,
    end_straight_length: float = 0.01,
    **routing_params,
) -> List[Route]:
    r"""Returns list of routes.

    Args:
        ports1: list of start ports.
        ports2: list of end ports.
        route_filter: filter to apply to the manhattan waypoints
            e.g `get_route_from_waypoints` for deep etch strip straight
        separation: center to center waveguide spacing.
        extension_length: in um.
        start_straight_length: extends in um.
        end_straight_length: in um.

    Returns:
        list of routes, where each route has references, ports and length.


    Used for routing multiple ports back to a bundled input in a component


    ::

        X: start ports
        D: End ports


                                    X------
                                    X----  |
                                   ...   | |
                                    X--\ | |
                                       | | |
            /--------------------------/ | |
            |                            | |
            | /--------------------------/ |
            | |                            |
            | | /--------------------------/
            | | |
            | | \--D
            | \----D
            |     ...
            \------D
            /------D
            |     ...
            | /----D
            | | /--D
            | | |
            | | \--------------------------
            | |                            |
            | \--------------------------\ |
            |                            | |
            \--------------------------\ | |
                                       | | |
                                    X--/ | |
                                   ...   | |
                                    X----/ |
                                    X------/
        '''

    """
    routes = _get_bundle_uindirect_waypoints(
        ports1,
        ports2,
        separation=separation,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        routing_func=generate_manhattan_waypoints,
        extension_length=extension_length,
        **routing_params,
    )

    return [route_filter(route, **routing_params) for route in routes]


def _get_bundle_uindirect_waypoints(
    ports1: List[Port],
    ports2: List[Port],
    routing_func: Callable = generate_manhattan_waypoints,
    separation: float = 5.0,
    extension_length: float = 0.0,
    start_straight_length: float = 0.01,
    end_straight_length: float = 0.01,
    **routing_func_params,
) -> List[ndarray]:

    nb_ports = len(ports1)

    for p in ports1:
        p.orientation = (
            p.orientation % 360 if p.orientation is not None else p.orientation
        )

    for p in ports2:
        p.orientation = (
            p.orientation % 360 if p.orientation is not None else p.orientation
        )

    if len(ports2) != nb_ports:
        raise ValueError(
            "Number of start ports should match number of end ports."
            "Got {} {}".format(len(ports1), len(ports2))
        )

    if len({p.orientation for p in ports1}) > 1:
        raise ValueError(f"All start port angles should be the same. Got {ports1}")

    if len({p.orientation for p in ports2}) > 1:
        raise ValueError(f"All end port angles should be the same. Got {ports2}")

    xs_end = [p.x for p in ports2]
    ys_end = [p.y for p in ports2]

    # Compute the bundle axis
    if ports1[0].orientation in [0, 180]:
        axis = "X"
    else:
        axis = "Y"

    # Split start ports in two groups:
    #    - the ones on the south/west of end ports (depending on bundle axis)
    #    - the ones on the north/east of end ports (depending on bundle axis)

    if axis == "X":
        y_cut = 0.5 * (min(ys_end) + max(ys_end))
        group1 = [p for p in ports1 if p.y <= y_cut]
        group2 = [p for p in ports1 if p.y > y_cut]

        if ports1[0].orientation == 0 and ports2[0].orientation == 180:
            """X->

            <-D
                 X->

            """
            # To go back to a U bundle
            group1_route_directives = ["north", "west"]
            group2_route_directives = ["south", "west"]

        elif ports1[0].orientation == 180 and ports2[0].orientation == 0:
            """
            <-X
                 D->
            <-X
            """
            # To go back to a U bundle
            group1_route_directives = ["north", "east"]
            group2_route_directives = ["south", "east"]

        else:
            print("u_undirect_bundle not designed to work in this case")

    if axis == "Y":
        x_cut = 0.5 * (min(xs_end) + max(xs_end))
        group1 = [p for p in ports1 if p.x <= x_cut]
        group2 = [p for p in ports1 if p.x > x_cut]

        if ports1[0].orientation == 90 and ports2[0].orientation == 270:
            """

            ^     ^
            |     |
            X     X
               D
               |

            """
            # To go back to a U bundle
            group1_route_directives = ["east", "south"]
            group2_route_directives = ["west", "south"]

        elif ports1[0].orientation == 270 and ports2[0].orientation == 90:
            """
               ^
               |
               D
            X     X
            |     |

            """
            # To go back to a U bundle
            group1_route_directives = ["east", "north"]
            group2_route_directives = ["west", "north"]

        else:
            print("u_undirect_bundle not designed to work in this case")

    # Do the routing directives to get back to a u_bundle direct case

    routing_param = {
        "routing_func": routing_func,
        "separation": separation,
        **routing_func_params,
    }

    # Multiple sections of different routes are generated in different places.
    # At the output, we want a list of routes. (not a list of portions of route)
    # dict_connections keeps track of these sections

    dict_connections = {i: [] for i in range(nb_ports)}

    def add_connections(conns) -> None:
        """Adds connections.

        Ensures that each section in a batch of connection. is added to
        the correct route. Also we don't know in which order the routes
        are given (from beginning to end or vice versa)

        """
        end_prev_conns = [(k, v[-1][-1]) for k, v in dict_connections.items()]
        for c in conns:
            p = c[0]
            for i, q in end_prev_conns:
                if np.abs(p - q).sum() < 1e-9:
                    dict_connections[i] += [c]
                    break

    # First part
    # print(group1_route_directives)
    conn1, tmp_ports1 = route_ports_to_side(
        group1,
        group1_route_directives[0],
        extension_length=extension_length,
        **routing_param,
    )

    conn2, tmp_ports2 = route_ports_to_side(
        group2,
        group2_route_directives[0],
        extension_length=extension_length,
        **routing_param,
    )
    conn = conn1 + conn2
    dict_connections = {i: [c] for i, c in enumerate(conn)}

    # Second part
    conn1, tmp_ports1 = route_ports_to_side(
        tmp_ports1, group1_route_directives[1], **routing_param
    )

    conn2, tmp_ports2 = route_ports_to_side(
        tmp_ports2, group2_route_directives[1], **routing_param
    )

    add_connections(conn1 + conn2)

    bundle_params = {
        **routing_param,
        "start_straight_length": start_straight_length,
        "end_straight_length": end_straight_length,
    }

    ports2.sort(key=lambda p: p.y)
    conns = []
    if tmp_ports1:
        conn1 = _get_bundle_udirect_waypoints(
            tmp_ports1, ports2[: len(tmp_ports1)], **bundle_params
        )
        conns.append(conn1)
    if tmp_ports2:
        conn2 = _get_bundle_udirect_waypoints(
            tmp_ports2, ports2[len(tmp_ports1) :], **bundle_params
        )
        conns.append(conn2)
    if len(conns) > 1:
        add_connections(conn1 + conn2)
    elif len(conns) == 1:
        add_connections(conns[0])
    else:
        raise ValueError("No connections generated!")

    def _merge_connections(list_of_points):

        a = [list_of_points[0]]
        a += [point[1:] for point in list_of_points[1:]]
        b = np.vstack(a)
        b = remove_identicals(b)
        b = remove_flat_angles(b)
        return b

    connections = [_merge_connections(c) for c in dict_connections.values()]
    return connections



import numpy as np


def coupling_length(
    neff1: float,
    neff2: float,
    power_ratio: float = 1.0,
    wavelength: float = 1.55,
) -> float:
    """Returns directional coupler coupling length (um) to achieve power_ratio.

    Args:
        neff1: even supermode of the directional coupler.
        neff2: odd supermode of the directional coupler.
        power_ratio: p2/p1, where 1 means 100% power transfer.
        wavelength: in um.
    """
    dneff = (neff1 - neff2).real
    return wavelength / (np.pi * dneff) * np.arcsin(np.sqrt(power_ratio))



"""Serialize component settings into YAML or strings."""

import copy
import functools
import hashlib
import inspect
import pathlib
from typing import Any, Dict

import numpy as np
import orjson
import pydantic
import toolz
from omegaconf import DictConfig, OmegaConf
from phidl.device_layout import Path as PathPhidl


def clean_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """Cleans dictionary recursively."""
    for k, v in d.items():
        d[k] = clean_dict(dict(v)) if isinstance(v, dict) else clean_value_json(v)
    return d


def get_string(value: Any) -> str:
    try:
        s = orjson.dumps(
            value, option=orjson.OPT_SERIALIZE_NUMPY, default=clean_value_name
        ).decode()
    except TypeError as e:
        print(f"Error serializing {value!r}")
        raise e
    return s


def clean_value_json(value: Any) -> Any:
    """Return JSON serializable object."""
    if isinstance(value, pydantic.BaseModel):
        return value.dict()

    elif isinstance(value, bool):
        return value

    elif isinstance(value, (np.integer, int)):
        return int(value)

    elif isinstance(value, (float, np.inexact, np.float64)):
        return float(value)

    elif isinstance(value, np.ndarray):
        value = np.round(value, 3)
        return orjson.dumps(value, option=orjson.OPT_SERIALIZE_NUMPY).decode()
    elif callable(value) and isinstance(value, functools.partial):
        sig = inspect.signature(value.func)
        args_as_kwargs = dict(zip(sig.parameters.keys(), value.args))
        args_as_kwargs.update(**value.keywords)
        clean_dict(args_as_kwargs)
        args_as_kwargs.pop("function", None)

        func = value.func
        while hasattr(func, "func"):
            func = func.func
        return dict(function=func.__name__, settings=args_as_kwargs)

    elif hasattr(value, "to_dict"):
        return value.to_dict()
    elif callable(value) and isinstance(value, toolz.functoolz.Compose):
        value = [clean_value_json(value.first)] + [
            clean_value_json(func) for func in value.funcs
        ]
    elif callable(value) and hasattr(value, "__name__"):
        value = dict(function=value.__name__)
    elif isinstance(value, PathPhidl):
        value = value.hash_geometry()
    elif isinstance(value, pathlib.Path):
        value = value.stem
    elif isinstance(value, dict):
        value = copy.deepcopy(value)
        value = clean_dict(value)
    elif isinstance(value, DictConfig):
        value = clean_dict(OmegaConf.to_container(value))
    else:
        try:
            value_json = orjson.dumps(
                value, option=orjson.OPT_SERIALIZE_NUMPY, default=clean_value_json
            )
            value = orjson.loads(value_json)
        except TypeError as e:
            print(f"Error serializing {value!r}")
            raise e
    return value

    # elif isinstance(value, (tuple, list, ListConfig)):
    #     value = [clean_value_json(i) for i in value]
    # elif value is None:
    #     value = None
    # elif hasattr(value, "name"):
    #     value = value.name
    # elif hasattr(value, "get_name"):
    #     value = value.get_name()
    # else:
    #     value = str(value)


def clean_value_name(value: Any) -> str:
    """Returns a string representation of an object."""
    # value1 = clean_value_json(value)
    # print(type(value), value, value1, str(value1))
    return str(clean_value_json(value))


def get_hash(value: Any) -> str:
    return hashlib.md5((clean_value_name(value)).encode()).hexdigest()[:8]


if __name__ == "__main__":
    import gdsfactory as gf

    # f = gf.partial(gf.c.straight, length=3)
    # d = clean_value_json(f)
    # print(f"{d!r}")

    f = gf.partial(gf.c.straight, length=3)
    c = f()
    d = clean_value_json(c)
    print(d, d)



"""tidy3d mode solver.

tidy3d has a powerful open source mode solver.

tidy3d can:

- compute bend modes.
- compute mode overlaps.

TODO:
- calculate dispersion

Maybe:

- combine modes package (based on modesolverpy), MPB and tidy3d APIs

"""

import pathlib
from types import SimpleNamespace
from typing import Callable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors
from pydantic import BaseModel, Extra
from scipy.constants import c as SPEED_OF_LIGHT
from tidy3d.plugins.mode.solver import compute_modes
from tqdm.auto import tqdm
from typing_extensions import Literal

from gdsfactory.config import CONFIG, logger
from gdsfactory.serialization import get_hash
from gdsfactory.simulation.gtidy3d.materials import si, sin, sio2
from gdsfactory.types import PathType

nm = 1e-3


def plot(
    X,
    Y,
    n,
    mode=None,
    num_levels: int = 8,
    n_cmap=None,
    mode_cmap=None,
    axes="xy",
    title=None,
    normalize_mode=False,
) -> None:
    """Plot waveguide index or mode in matplotlib.

    Args:
        X: x array.
        Y: y array.
        n: refractive index.
        mode: mode number.
        num_levels: for plot.
        n_cmap: refractive index color map.
        mode_cmap: mode color map.
        axes: "xy".
        title: for the plot.
        normalize_mode: divide by maximum value.

    """
    x, y = axes
    if n_cmap is None:
        n_cmap = colors.LinearSegmentedColormap.from_list(
            name="custom", colors=["#ffffff", "#c1d9ed"]
        )
    if mode_cmap is None:
        mode_cmap = "inferno"
    if mode is not None:
        mode = np.abs(mode)
        if normalize_mode:
            mode = mode / mode.max()
        plt.contour(
            X,
            Y,
            mode,
            cmap=mode_cmap,
            levels=np.linspace(mode.min(), mode.max(), num_levels),
        )
        plt.colorbar(label="mode")
    n = np.real(n)
    plt.pcolormesh(X, Y, n, cmap=n_cmap)
    plt.colorbar(label="n")
    plt.xlabel(x)
    plt.ylabel(y)
    plt.grid(True, alpha=0.4)
    plt.axis("scaled")
    if title is not None:
        plt.title(title)


def _mesh_1d(x_min, x_max, mesh_x):
    if isinstance(mesh_x, int):
        x = np.linspace(x_min, x_max, mesh_x + 1)
        dx = x[1] - x[0]
    elif isinstance(mesh_x, float):
        dx = mesh_x
        x = np.arange(x_min, x_max + 0.999 * dx, dx)
        mesh_x = x.shape[0]
    else:
        raise ValueError("Invalid 'mesh_x'")
    return x, mesh_x, dx


def create_mesh(x_min, y_min, x_max, y_max, mesh_x, mesh_y):
    x, mesh_x, dx = _mesh_1d(x_min, x_max, mesh_x)
    y, mesh_y, dy = _mesh_1d(y_min, y_max, mesh_y)

    Yx, Xx = np.meshgrid(y[:-1], x[:-1] + dx / 2)
    Yy, Xy = np.meshgrid(y[:-1] + dy / 2, x[:-1])
    Yz, Xz = np.meshgrid(y[:-1], x[:-1])
    return x, y, Xx, Yx, Xy, Yy, Xz, Yz


SETTINGS = [
    "wavelength",
    "wg_width",
    "wg_thickness",
    "ncore",
    "nclad",
    "slab_thickness",
    "t_box",
    "t_clad",
    "xmargin",
    "resolution",
    "nmodes",
    "bend_radius",
]

SETTINGS_COUPLER = set(SETTINGS + ["wg_width1", "wg_width2", "gap"])
Precision = Literal["single", "double"]
FilterPol = Literal["te", "tm"]


class Waveguide(BaseModel):
    """Waveguide Model.

    Parameters:
        wavelength: (um).
        wg_width: waveguide width.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
        cache: filepath for caching modes. If None does not use file cache.
        precision: single or double.
        filter_pol: te, tm or None.

    ::

          __________________________
          |
          |
          |         width     xmargin
          |     <----------> <------>
          |      ___________   _ _ _
          |     |           |       |
          |_____|  ncore    |_______|
          |                         | wg_thickness
          |slab_thickness    nslab  |
          |_________________________|
          |
          |        nclad
          |__________________________
          <------------------------>
                   w_sim

    """

    wavelength: float
    wg_width: float
    wg_thickness: float
    ncore: Union[float, Callable[[str], float]]
    nclad: Union[float, Callable[[str], float]]
    slab_thickness: float
    t_box: float = 2.0
    t_clad: float = 2.0
    xmargin: float = 1.0
    resolution: int = 100
    nmodes: int = 4
    bend_radius: Optional[float] = None
    cache: Optional[PathType] = CONFIG["modes"]
    precision: Precision = "single"
    filter_pol: Optional[FilterPol] = None

    class Config:
        """Config for Waveguide."""

        extra = Extra.allow

    @property
    def t_sim(self):
        return self.t_box + self.wg_thickness + self.t_clad

    @property
    def settings(self):
        return SETTINGS

    @property
    def w_sim(self):
        return self.wg_width + 2 * self.xmargin

    @property
    def filepath(self) -> Optional[pathlib.Path]:
        if self.cache is None:
            return
        cache = pathlib.Path(self.cache)
        cache.mkdir(exist_ok=True, parents=True)
        settings = {setting: getattr(self, setting) for setting in self.settings}
        return cache / f"{get_hash(settings)}.npz"

    def get_ncore(self, wavelength: Optional[float] = None):
        wavelength = wavelength or self.wavelength
        return self.ncore(wavelength) if callable(self.ncore) else self.ncore

    def get_nclad(self, wavelength: Optional[float] = None):
        wavelength = wavelength or self.wavelength
        return self.nclad(wavelength) if callable(self.nclad) else self.nclad

    def get_n(self, Y, Z):
        """Return index matrix for a waveguide.

        Args:
            Y: 2D array.
            Z: 2D array.

        """
        w = self.wg_width
        ncore = self.get_ncore()
        nclad = self.get_nclad()
        t_box = self.t_box
        wg_thickness = self.wg_thickness
        slab_thickness = self.slab_thickness
        t_clad = self.t_clad

        n = np.ones_like(Y) * nclad
        n[
            (-w / 2 - t_clad / 2 <= Y)
            & (Y <= w / 2 + t_clad / 2)
            & (Z >= t_box)
            & (Z <= t_box + wg_thickness + t_clad)
        ] = nclad
        n[(Z <= 1.0 + slab_thickness + t_clad)] = nclad
        n[
            (-w / 2 <= Y) & (Y <= w / 2) & (Z >= t_box) & (Z <= t_box + wg_thickness)
        ] = ncore
        n[(Z >= t_box) & (Z <= t_box + slab_thickness)] = (
            ncore if slab_thickness else nclad
        )
        return n

    def plot_index(self) -> None:
        x, y, Xx, Yx, Xy, Yy, Xz, Yz = create_mesh(
            -self.w_sim / 2,
            0.0,
            +self.w_sim / 2,
            self.t_sim,
            self.resolution,
            self.resolution,
        )

        nx = self.get_n(
            Xx,
            Yx,
        )
        plot(Xx, Yx, nx)
        plt.show()

    def compute_modes(
        self,
        overwrite: bool = False,
        with_fields: bool = True,
    ) -> None:
        """Compute modes.

        Args:
            overwrite: overwrite file cache.
            with_fields: include field data.
        """
        if hasattr(self, "neffs") and not overwrite:
            return

        wavelength = self.wavelength
        x, y, Xx, Yx, Xy, Yy, Xz, Yz = create_mesh(
            -self.w_sim / 2,
            0.0,
            +self.w_sim / 2,
            self.t_sim,
            self.resolution,
            self.resolution,
        )

        nx = self.get_n(
            Xx,
            Yx,
        )
        ny = self.get_n(
            Xy,
            Yy,
        )
        nz = self.get_n(
            Xz,
            Yz,
        )
        self.nx, self.ny, self.nz = nx, ny, nz
        self.Xx, self.Yx, self.Xy, self.Yy, self.Xz, self.Yz = Xx, Yx, Xy, Yy, Xz, Yz

        if self.cache and self.filepath and self.filepath.exists():
            data = np.load(self.filepath)

            if with_fields:
                self.Ex = data["Ex"]
                self.Ey = data["Ey"]
                self.Ez = data["Ez"]
                self.Hx = data["Hx"]
                self.Hy = data["Hy"]
                self.Hz = data["Hz"]
            self.neffs = data["neffs"]
            logger.info(f"load {self.filepath} mode data from file cache.")
            return

        ((Ex, Ey, Ez), (Hx, Hy, Hz)), neffs = (
            x.squeeze()
            for x in compute_modes(
                eps_cross=[nx**2, ny**2, nz**2],
                coords=[x, y],
                freq=SPEED_OF_LIGHT / (wavelength * 1e-6),
                mode_spec=SimpleNamespace(
                    num_modes=self.nmodes,
                    bend_radius=self.bend_radius,
                    bend_axis=1,
                    angle_theta=0.0,
                    angle_phi=0.0,
                    num_pml=(0, 0),
                    target_neff=self.get_ncore(wavelength),
                    sort_by="largest_neff",
                    precision=self.precision,
                    filter_pol=self.filter_pol,
                ),
            )
        )

        self.Ex, self.Ey, self.Ez = Ex, Ey, Ez
        self.Hx, self.Hy, self.Hz = Hx, Hy, Hz
        self.neffs = neffs

        if with_fields:
            data = dict(
                Ex=self.Ex,
                Ey=self.Ey,
                Ez=self.Ez,
                Hx=self.Hx,
                Hy=self.Hy,
                Hz=self.Hz,
                neffs=self.neffs,
            )
        else:
            data = dict(neffs=self.neffs)
        if self.filepath:
            np.savez_compressed(self.filepath, **data)
            logger.info(f"write {self.filepath} mode data to file cache.")

    def compute_mode_properties(self) -> Tuple[List[float], List[float], List[float]]:
        """Computes mode areas, fraction_te and fraction_tm."""
        if not hasattr(self, "neffs"):
            self.compute_modes()
        mode_areas = []
        fraction_te = []
        fraction_tm = []

        for mode_index in range(self.nmodes):
            e_fields = (
                self.Ex[..., mode_index],
                self.Ey[..., mode_index],
                self.Ez[..., mode_index],
            )
            h_fields = (
                self.Hx[..., mode_index],
                self.Hy[..., mode_index],
                self.Hz[..., mode_index],
            )

            areas_e = [np.sum(np.abs(e) ** 2) for e in e_fields]
            areas_e /= np.sum(areas_e)
            areas_e *= 100

            areas_h = [np.sum(np.abs(h) ** 2) for h in h_fields]
            areas_h /= np.sum(areas_h)
            areas_h *= 100

            fraction_te.append(areas_e[0] / (areas_e[0] + areas_e[1]))
            fraction_tm.append(areas_e[1] / (areas_e[0] + areas_e[1]))

            areas = areas_e.tolist()
            areas.extend(areas_h)
            mode_areas.append(areas)

        self.mode_areas = mode_areas
        self.fraction_te = fraction_te
        self.fraction_tm = fraction_tm
        return mode_areas, fraction_te, fraction_tm

    def plot_Ex(self, mode_index: int = 0) -> None:
        if not hasattr(self, "neffs"):
            self.compute_modes()

        nx, neffs, Ex = self.nx, self.neffs, self.Ex
        neff_, Ex_ = np.real(neffs[mode_index]), Ex[..., mode_index]
        plot(self.Xx, self.Yx, nx, mode=np.abs(Ex_) ** 2, title=f"Ex::{neff_:.3f}")
        plt.show()

    def plot_Ey(self, mode_index: int = 0) -> None:
        if not hasattr(self, "neffs"):
            self.compute_modes()

        nx, neffs, Ey = self.nx, self.neffs, self.Ey
        neff_, Ey_ = np.real(neffs[mode_index]), Ey[..., mode_index]
        plot(self.Xx, self.Yx, nx, mode=np.abs(Ey_) ** 2, title=f"Ey::{neff_:.3f}")
        plt.show()

    def _repr_html_(self) -> str:
        """Show index in matplotlib for jupyter notebooks."""
        self.plot_index()
        return self.__repr__()

    def __repr__(self) -> str:
        """Show waveguide name."""
        return ", \n".join([f"{k} = {getattr(self, k)!r}" for k in self.settings])

    def get_overlap(
        self, wg: "Waveguide", mode_index1: int = 0, mode_index2: int = 0
    ) -> float:
        """Returns mode overlap integral.

        Args:
            wg: other waveguide.

        """
        wg1 = self
        wg2 = wg

        wg1.compute_modes()
        wg2.compute_modes()

        return np.sum(
            np.conj(wg1.Ex[..., mode_index1]) * wg2.Hy[..., mode_index2]
            - np.conj(wg1.Ey[..., mode_index1]) * wg2.Hx[..., mode_index2]
            + wg2.Ex[..., mode_index2] * np.conj(wg1.Hy[..., mode_index1])
            - wg2.Ey[..., mode_index2] * np.conj(wg1.Hx[..., mode_index1])
        )


class WaveguideCoupler(Waveguide):
    """Waveguide coupler Model.

    Parameters:
        wavelength: (um).
        wg_width1: left waveguide width in um.
        wg_width2: right waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
        cache: filepath for caching modes. If None does not use file cache.

    ::

        -w1-gap/2

            wg_width1     wg_width2
            <------->     <------->
             _______   |   _______   __
            |       |  |  |       | |
            |       |  |  |       | |
            |       |_____|       | | wg_thickness
            |slab_thickness       | |
            |_____________________| |__
                    <----->
                      gap
            <--------------------->
                     w_sim

    """

    wg_width: Optional[float] = None
    wg_width1: float
    wg_width2: float
    gap: float

    @property
    def w_sim(self):
        return self.wg_width1 + self.wg_width2 + self.gap + 2 * self.xmargin

    def get_n(self, Y, Z):
        """Return index matrix for a waveguide coupler.

        Args:
            Y: 2D array.
            Z: 2D array.

        """
        w1 = self.wg_width1
        w2 = self.wg_width2
        gap = self.gap
        ncore = self.get_ncore()
        nclad = self.get_nclad()
        t_box = self.t_box
        wg_thickness = self.wg_thickness
        slab_thickness = self.slab_thickness
        t_clad = self.t_clad

        n = np.ones_like(Y) * nclad
        n[(Z <= 1.0 + slab_thickness + t_clad)] = nclad
        n[
            (-w1 - gap / 2 <= Y)
            & (Y <= -gap / 2)
            & (Z >= t_box)
            & (Z <= t_box + wg_thickness)
        ] = ncore
        n[
            (gap / 2 <= Y)
            & (Y <= gap / 2 + w2)
            & (Z >= t_box)
            & (Z <= t_box + wg_thickness)
        ] = ncore
        n[(Z >= t_box) & (Z <= t_box + slab_thickness)] = (
            ncore if slab_thickness else nclad
        )
        return n

    @property
    def settings(self):
        return SETTINGS_COUPLER

    def find_coupling(self, power_ratio: float = 1.0) -> float:
        """Returns the coupling length (um) of the directional coupler to achieve power_ratio, where 1 means 100% power transfer."""
        if not hasattr(self, "neffs"):
            self.compute_modes()
        neff1 = self.neffs[0]
        neff2 = self.neffs[1]
        dneff = (neff1 - neff2).real
        return self.wavelength / (np.pi * dneff) * np.arcsin(np.sqrt(power_ratio))


def sweep_bend_loss(
    bend_radius_min: float = 2.0,
    bend_radius_max: float = 5,
    steps: int = 4,
    mode_index: int = 0,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """Returns overlap integral squared for the bend mode mismatch loss.

    The loss is squared because you hit the bend loss twice
    (from bend to straight and from straight to bend).

    Args:
        bend_radius_min: min bend radius (um).
        bend_radius_max: max bend radius (um).
        steps: number of steps.
        mode_index: where 0 is the fundamental mode.

    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width in um.
        wg_thickness: thickness waveguide (um).
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.

    """
    r = np.linspace(bend_radius_min, bend_radius_max, steps)
    integral = np.zeros_like(r)

    wg = Waveguide(**kwargs)

    for i, radius in tqdm(enumerate(r)):
        wg_bent = Waveguide(bend_radius=radius, **kwargs)
        wg.get_overlap(wg_bent, mode_index1=mode_index, mode_index2=mode_index)

        # normalized overlap integral
        integral[i] = np.abs(
            wg.get_overlap(wg_bent, mode_index, mode_index) ** 2
            / wg.get_overlap(wg, mode_index, mode_index)
            / wg.get_overlap(wg_bent, mode_index, mode_index)
        )

    return r, integral**2


def sweep_width(
    width1: float = 200 * nm,
    width2: float = 1000 * nm,
    steps: int = 12,
    nmodes: int = 4,
    **kwargs,
) -> pd.DataFrame:
    """Sweep waveguide width and compute effective index.

    Args:
        width1: starting waveguide width in um.
        width2: end waveguide width in um.
        steps: number of points.
        nmodes: number of modes to compute.


    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).

    """
    width = np.linspace(width1, width2, steps)
    neff = {mode_number: [] for mode_number in range(1, nmodes + 1)}
    for wg_width in tqdm(width):
        wg = Waveguide(nmodes=nmodes, wg_width=wg_width, **kwargs)
        wg.compute_modes()
        for mode_number in range(1, nmodes + 1):
            neff[mode_number].append(np.real(wg.neffs[mode_number]))

    df = pd.DataFrame(neff)
    df["width"] = width
    return df


def group_index(
    wavelength: float, wavelength_step: float = 0.01, mode_index: int = 0, **kwargs
) -> float:
    """Returns group_index.

    Args:
        wavelength: (um).
        wavelength_step: in um.
        mode_index: integer.

    Keyword Args:
        wg_width: waveguide width.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).

    """
    wc = Waveguide(wavelength=wavelength, **kwargs)
    wf = Waveguide(
        wavelength=wavelength + wavelength_step,
        **kwargs,
    )
    wb = Waveguide(
        wavelength=wavelength - wavelength_step,
        **kwargs,
    )
    wc.compute_modes()
    wb.compute_modes()
    wf.compute_modes()

    nc = np.real(wc.neffs[mode_index])
    nb = np.real(wb.neffs[mode_index])
    nf = np.real(wf.neffs[mode_index])
    return nc - wavelength * (nf - nb) / (2 * wavelength_step)


def plot_sweep_width(
    width1: float = 200 * nm,
    width2: float = 1000 * nm,
    steps: int = 12,
    nmodes: int = 4,
    cmap: str = "magma",
    **kwargs,
) -> None:
    """Sweep waveguide width and compute effective index.

    Args:
        width1: starting waveguide width in um.
        width2: end waveguide width in um.
        steps: number of points.
        nmodes: number of modes to compute.
        cmap: colormap for the TE fraction.

    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width in um.
        wg_thickness: thickness waveguide (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        slab_thickness: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        xmargin: margin from waveguide edge to each side (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).

    """
    width = np.linspace(width1, width2, steps)
    neff = {mode_number: [] for mode_number in range(nmodes)}

    for wg_width in tqdm(width):
        wg = Waveguide(nmodes=nmodes, wg_width=wg_width, **kwargs)
        wg.compute_modes()
        wg.compute_mode_properties()
        for mode_number in range(nmodes):
            n = np.real(wg.neffs[mode_number])
            neff[mode_number].append(n)

            fraction_te = wg.fraction_te[mode_number]
            plt.scatter(wg_width, n, c=fraction_te, vmin=0, vmax=1, cmap=cmap)

    for mode_number in range(nmodes):
        plt.plot(width, neff[mode_number], c="gray")

    plt.colorbar().set_label("TE-Fraction")
    plt.xlabel("width (um)")
    plt.ylabel("neff")


__all__ = (
    "Waveguide",
    "plot_sweep_width",
    "si",
    "sin",
    "sio2",
    "sweep_bend_loss",
    "sweep_width",
    "group_index",
)

if __name__ == "__main__":
    c = Waveguide(
        wavelength=1.55,
        wg_width=500 * nm,
        wg_thickness=220 * nm,
        slab_thickness=0 * nm,
        ncore=si,
        nclad=sio2,
    )
    c = WaveguideCoupler(
        wavelength=1.55,
        wg_width1=500 * nm,
        wg_width2=500 * nm,
        gap=200 * nm,
        wg_thickness=220 * nm,
        slab_thickness=100 * nm,
        ncore=si,
        nclad=sio2,
    )
    print(c.find_coupling())
    # c.plot_index()

    # mode_areas, te, tm = c.compute_mode_properties()
    # c.plot_Ex(0)
    # nitride = find_modes(wavelength=1.55, wg_width=1.0, wg_thickness=0.4, ncore=2.0)
    # nitride.plot_index()

    # c = pickle_load("strip.pkl")

    # c0 = Waveguide(slab_thickness=0)
    # c0.plot_Ex(index=0)
    # c0.pickle_dump("strip.pkl")

    # c1 = Waveguide(slab_thickness=0, bend_radius=5)
    # c1.plot_Ex()
    # c1.pickle_dump("strip_bend5.pkl")

    # c = Waveguide(slab_thickness=90e-3, bend_radius=5)
    # c.plot_index()

    # r, integral = sweep_bend_loss(
    #     wavelength=1.55,
    #     wg_width=0.5,
    #     wg_thickness=0.22,
    #     slab_thickness=0.0,
    #     ncore=si,
    #     nclad=sio2,
    # )
    # plt.plot(r, integral / max(integral), ".")
    # plt.xlabel("bend radius (um)")
    # plt.show()

    # rib = find_modes(
    #     wavelength=1.55, wg_width=0.5, wg_thickness=0.22, slab_thickness=0.15, ncore=3.4, nclad=1.44
    # )
    # nitride = find_modes(
    #     wavelength=1.55,
    #     wg_width=1.0,
    #     wg_thickness=0.4,
    #     ncore=2.0,
    #     nclad=sio2,
    # )

    # nitride.plot_index()
    # nitride.plot_Ex(index=0)

    # plot_sweep_width(
    #     steps=3,
    #     wavelength=1.55,
    #     wg_thickness=220 * nm,
    #     slab_thickness=0 * nm,
    #     ncore=si,
    #     nclad=sio2,
    # )
    # plt.show()
    # ng = group_index(
    #     wg_width=500 * nm,
    #     wavelength=1.55,
    #     wg_thickness=220 * nm,
    #     slab_thickness=0 * nm,
    #     ncore=si,
    #     nclad=sio2,
    # )
    # print(ng)
    # plt.show()



import warnings

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler180
from gdsfactory.types import ComponentSpec, CrossSectionSpec


@gf.cell
def delay_snake3(
    length: float = 1600.0,
    length0: float = 0.0,
    n: int = 2,
    bend180: ComponentSpec = bend_euler180,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    r"""Snake input facing west.

    Args:
        length: total length.
        length0: initial offset.
        n: number of loops.
        bend180: ubend spec.
        cross_section: cross_section spec.
        kwargs: cross_section settings.

    .. code::

                 | length0   |

                 >---------\
                            \bend180.info['length']
                            /
       |-------------------/
       |
       |------------------->

       |   delta_length    |


    """
    if n % 2:
        warnings.warn(f"rounding {n} to {n//2 *2}", stacklevel=3)
        n = n // 2 * 2
    bend180 = gf.get_component(bend180, cross_section=cross_section, **kwargs)

    delta_length = (length - length0 - n * bend180.info["length"]) / (n + 1)
    assert (
        delta_length > 0
    ), "Snake is too short: either reduce length0, increase the total length,\
    or decrease n"

    s0 = gf.components.straight(cross_section=cross_section, length=length0, **kwargs)
    sd = gf.components.straight(
        cross_section=cross_section, length=delta_length, **kwargs
    )

    symbol_to_component = {
        "_": (s0, "o1", "o2"),
        "-": (sd, "o1", "o2"),
        ")": (bend180, "o2", "o1"),
        "(": (bend180, "o1", "o2"),
    }

    sequence = "_)" + n // 2 * "-(-)"
    sequence = sequence[:-1]
    return gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component
    )


if __name__ == "__main__":
    # c = test_delay_snake3_length()
    length = 200.0
    c = delay_snake3(n=2, length=length, length0=50)
    c.show(show_ports=True)



import numpy as np
import pytest

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.routing.manhattan import RouteWarning, route_manhattan

TOLERANCE = 0.001
DEG2RAD = np.pi / 180
RAD2DEG = 1 / DEG2RAD

O2D = {0: "East", 180: "West", 90: "North", 270: "South"}


@cell
def test_manhattan_sbend_pass() -> Component:
    c = gf.Component("demo_sbend")
    length = 10
    c1 = c << gf.components.straight(length=length)
    c2 = c << gf.components.straight(length=length)

    dy = 4.0
    c2.y = dy
    c2.movex(length + 20)

    route = route_manhattan(
        input_port=c1.ports["o2"],
        output_port=c2.ports["o1"],
        radius=5.0,
        with_sbend=True,
    )

    c.add(route.references)
    if route.labels:
        c.add(route.labels)
    return c


@cell
def test_manhattan_sbend_fail() -> Component:
    with pytest.warns(RouteWarning):
        c = gf.Component("demo_sbend")
        length = 10
        c1 = c << gf.components.straight(length=length)
        c2 = c << gf.components.straight(length=length)

        dy = 4.0
        c2.y = dy
        c2.movex(length + 20)

        route = route_manhattan(
            input_port=c1.ports["o2"],
            output_port=c2.ports["o1"],
            radius=5.0,
            with_sbend=False,
        )

        c.add(route.references)
        if route.labels:
            c.add(route.labels)
    return c


if __name__ == "__main__":
    # c = test_manhattan_sbend_pass()
    # c = test_manhattan_sbend_fail()
    # c.show(show_ports=True)

    c = gf.Component("demo_sbend")
    length = 10
    c1 = c << gf.components.straight(length=length)
    c2 = c << gf.components.straight(length=length)

    dy = 20.0
    dy = 4.0
    c2.y = dy
    c2.movex(length + 20)

    route = route_manhattan(
        input_port=c1.ports["o2"],
        output_port=c2.ports["o1"],
        radius=5.0,
        with_sbend=False,
    )

    c.add(route.references)
    c.show()



"""GDS regression test. Adapted from lytest.

TODO: adapt it into pytest_regressions

from pytest_regressions.file_regression import FileRegressionFixture

class GdsRegressionFixture(FileRegressionFixture):
    def check(self,
        contents,
        extension=".gds",
        basename=None,
        fullpath=None,
        binary=False,
        obtained_filename=None,
        check_fn=None,
            ):
        try:
            difftest(c)

"""
import filecmp
import pathlib
import shutil
from typing import Optional

from gdsfactory.component import Component
from gdsfactory.config import CONFIG, logger
from gdsfactory.gdsdiff.gdsdiff import gdsdiff


def difftest(
    component: Component,
    test_name: Optional[str] = None,
    xor: bool = False,
    dirpath: pathlib.Path = CONFIG["gdsdiff"],
) -> None:
    """Avoids GDS regressions tests on the GeometryDifference.

    If files are the same it returns None. If files are different runs XOR
    between new component and the GDS reference stored in dirpath and
    raises GeometryDifference if there are differences and show differences in klayout.

    If it runs for the fist time it just stores the GDS reference.


    Args:
        component: to test if it has changed.
        test_name: used to store the GDS file.
        xor: runs xor if there is difference.
        dirpath: defaults to cwd refers to where the test is being invoked.

    """
    from lytest.kdb_xor import GeometryDifference, run_xor

    # containers function_name is different from component.name
    # we store the container with a different name from original component
    test_name = test_name or (
        f"{component.function_name}_{component.name}"
        if hasattr(component, "function_name")
        and component.name != component.function_name
        else f"{component.name}"
    )
    filename = f"{test_name}.gds"
    ref_file = dirpath / "gds_ref" / filename
    run_file = dirpath / "gds_run" / filename
    diff_file = dirpath / "gds_diff" / filename

    component.write_gds(gdspath=run_file)

    if not ref_file.exists():
        component.write_gds(gdspath=ref_file)
        raise AssertionError(
            f"Reference GDS file for {test_name!r} not found. Writing to {ref_file!r}"
        )

    if filecmp.cmp(ref_file, run_file, shallow=False):
        return

    # component_reference = gf.import_gds(ref_file)
    # if component.hash_geometry() == component_reference.hash_geometry():
    #     print("same hash")
    #     return

    try:
        run_xor(str(ref_file), str(run_file), tolerance=1, verbose=False)
    except GeometryDifference as error:
        logger.error(error)
        diff = gdsdiff(ref_file, run_file, name=test_name, xor=xor)
        diff.write_gds(diff_file)
        diff.show(show_ports=False)
        print(
            f"\ngds_run {filename!r} changed from gds_ref {str(ref_file)!r}\n"
            "You can check the differences in Klayout GUI or run XOR with\n"
            f"gf gds diff --xor {ref_file} {run_file}\n"
        )

        try:
            val = input(
                "Would you like to save current GDS as the new reference? [Y/n] "
            )
            if val.upper().startswith("N"):
                raise
            logger.info(f"deleting file {str(ref_file)!r}")
            ref_file.unlink()
            shutil.copy(run_file, ref_file)
            raise
        except OSError as exc:
            raise GeometryDifference(
                "\n"
                f"{filename!r} changed from reference {str(ref_file)!r}\n"
                "To step over each error you can run `pytest -s`\n"
                "So you can check the differences in Klayout GUI\n"
            ) from exc


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    difftest(c)
    # test_component(c, None, None)



import toolz
from omegaconf import OmegaConf

import gdsfactory as gf


@gf.cell
def mask(size=(1000, 1000)):
    ring_te = toolz.compose(gf.routing.add_fiber_array, gf.components.ring_single)
    rings = gf.grid([ring_te(radius=r) for r in [10, 20, 50]])
    c = gf.Component()
    c << gf.components.die(size=size)
    c << rings
    return c


def test_mask_metadata() -> gf.Component:
    m = mask()
    gdspath = m.write_gds_with_metadata(gdspath="test_mask_metadata.gds")
    labels_path = gdspath.with_suffix(".csv")
    gf.labels.write_labels.write_labels_klayout(gdspath=gdspath, layer_label=(66, 0))

    yamlpath = gdspath.with_suffix(".yml")
    mask_metadata = OmegaConf.load(yamlpath)
    tm = gf.labels.merge_test_metadata(
        mask_metadata=mask_metadata, labels_path=labels_path
    )
    assert len(tm.keys()) == 3
    return m


if __name__ == "__main__":
    m = test_mask_metadata()
    m = mask()
    m.show()



"""read Sparameters from CSV file and returns sax model."""
import pathlib
from typing import Union

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from sax.typing_ import Float, Model
from typing_extensions import Literal

import gdsfactory as gf
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_lumerical,
    get_sparameters_path_meep,
    get_sparameters_path_tidy3d,
)

wl_cband = np.linspace(1.500, 1.600, 128)

PathType = Union[str, pathlib.Path]

Simulator = Literal["lumerical", "meep", "tidy3d"]


def model_from_npz(
    filepath: Union[PathType, np.ndarray],
    xkey: str = "wavelengths",
    xunits: float = 1,
    prefix: str = "s",
) -> Model:
    """Returns a SAX Sparameters Model from a npz file.

    The SAX Model is a function that returns a SAX SDict interpolated over wavelength.

    Args:
        filepath: CSV Sparameters path or pandas DataFrame.
        wl: wavelength to interpolate (um).
        xkey: key for wavelengths in file.
        xunits: x units in um from the loaded file (um). 1 means 1um.
        prefix: for the sparameters column names in file.

    """
    sp = filepath if isinstance(filepath, np.ndarray) else np.load(filepath)
    keys = list(sp.keys())

    if xkey not in keys:
        raise ValueError(f"{xkey!r} not in {keys}")

    x = jnp.asarray(sp[xkey] * xunits)
    wl = jnp.asarray(wl_cband)

    # make sure x is sorted from low to high
    idxs = jnp.argsort(x)
    x = x[idxs]
    sp = {k: v[idxs] for k, v in sp.items()}

    @jax.jit
    def model(wl: Float = wl):
        S = {}
        zero = jnp.zeros_like(x)

        for key in sp:
            if not key.startswith("wav"):
                port_mode0, port_mode1 = key.split(",")
                port0, mode0 = port_mode0.split("@")
                port1, mode1 = port_mode1.split("@")
                S[(port0, port1)] = jnp.interp(wl, x, sp.get(key, zero))

        return S

    return model


def model_from_csv(
    filepath: Union[PathType, pd.DataFrame],
    xkey: str = "wavelengths",
    xunits: float = 1,
    prefix: str = "s",
) -> Model:
    """Returns a SAX Sparameters Model from a CSV file.

    The SAX Model is a function that returns a SAX SDict interpolated over wavelength.

    Args:
        filepath: CSV Sparameters path or pandas DataFrame.
        wl: wavelength to interpolate (um).
        xkey: key for wavelengths in file.
        xunits: x units in um from the loaded file (um). 1 means 1um.
        prefix: for the sparameters column names in file.
    """
    df = filepath if isinstance(filepath, pd.DataFrame) else pd.read_csv(filepath)
    assert isinstance(df, pd.DataFrame)
    df = df.reset_index()  # maybe there is useful info in the index...
    dic = dict(zip(df.columns, jnp.asarray(df.values.T)))
    keys = list(dic.keys())

    if xkey not in keys:
        raise ValueError(f"{xkey!r} not in {keys}")

    nsparameters = (len(keys) - 1) // 2
    nports = int(nsparameters**0.5)

    x = jnp.asarray(dic[xkey] * xunits)
    wl = jnp.asarray(wl_cband)

    # make sure x is sorted from low to high
    idxs = jnp.argsort(x)
    x = x[idxs]
    dic = {k: v[idxs] for k, v in dic.items()}

    @jax.jit
    def model(wl: Float = wl):
        S = {}
        zero = jnp.zeros_like(x)
        for i in range(1, nports + 1):
            for j in range(1, nports + 1):
                m = jnp.interp(wl, x, dic.get(f"{prefix}{i}{j}m", zero))
                a = jnp.interp(wl, x, dic.get(f"{prefix}{i}{j}a", zero))
                S[f"o{i}", f"o{j}"] = m * jnp.exp(1j * a)

        return S

    return model


def _demo_mmi_lumerical_csv() -> None:
    import matplotlib.pyplot as plt
    from plot_model import plot_model

    filepath = get_sparameters_path_lumerical(gf.c.mmi1x2)
    mmi = model_from_csv(filepath=filepath, xkey="wavelengths")
    plot_model(mmi)
    plt.show()


def model_from_component(component, simulator: Simulator, **kwargs) -> Model:
    """Returns SAX model from lumerical FDTD simulations.

    Args:
        component: to simulate.
        simulator: meep, lumerical or tidy3d.
        kwargs: simulator settings.

    """
    simulators = ["lumerical", "meep", "tidy3d"]

    if simulator == "lumerical":
        filepath = get_sparameters_path_lumerical(component=component, **kwargs)
    elif simulator == "meep":
        filepath = get_sparameters_path_meep(component=component, **kwargs)
    elif simulator == "tidy3d":
        filepath = get_sparameters_path_tidy3d(component=component, **kwargs)
    else:
        raise ValueError(f"{simulator!r} no in {simulators}")
    return model_from_csv(filepath=filepath)


model_from_component_lumerical = gf.partial(model_from_component, simulator="fdtd")


if __name__ == "__main__":
    # mmi1x2 = gf.partial(model_from_component_lumerical, component=gf.components.mmi1x2)
    # mmi2x2 = gf.partial(model_from_component_lumerical, component=gf.components.mmi2x2)

    # grating_coupler_elliptical = model_from_csv(
    #     filepath=sparameters_path / "grating_coupler_ellipti_9d85a0c6_18c08cac.csv"
    # )

    # model_factory = dict(
    #     mmi1x2=mmi1x2, mmi2x2=mmi2x2, grating_coupler_elliptical=grating_coupler_elliptical
    # )
    # from gdsfactory.config import sparameters_path

    import matplotlib.pyplot as plt
    from plot_model import plot_model

    # s = from_csv(filepath=filepath, wavelength=wl_cband, xunits=1e-3)
    # s21 = s[("o1", "o3")]
    # plt.plot(wl_cband, np.abs(s21) ** 2)
    # plt.show()
    # filepath = get_sparameters_path_lumerical(gf.components.mmi1x2)
    # mmi = model_from_csv(filepath=filepath, xkey="wavelengths")
    # plot_model(mmi)
    # plt.show()
    # plot_model(grating_coupler_elliptical)
    # plt.show()
    # model = partial(
    #     from_csv, filepath=filepath, xunits=1, xkey="wavelengths", prefix="s"
    # )
    # sax.plot.plot_model(model)
    # plt.show()
    #  This looks correct
    # coupler_fdtd = gf.partial(
    #     sdict_from_csv,
    #     filepath=gf.config.sparameters_path / "coupler" / "coupler_G224n_L20_S220.csv",
    #     xkey="wavelength_nm",
    #     prefix="S",
    #     xunits=1e-3,
    # )
    # this looks wrong
    # coupler_fdtd = model_from_csv(
    #     filepath=sparameters_path / "coupler" / "coupler_G224n_L20_S220.csv",
    #     xkey="wavelength_nm",
    #     prefix="S",
    #     xunits=1e-3,
    # )
    # coupler_fdtd = model_from_npz(
    #     filepath=sparameters_path / "coupler" / "coupler_G224n_L20_S220.npz",
    #     xkey="wavelength_nm",
    #     prefix="S",
    #     xunits=1e-3,
    # )
    filepath = get_sparameters_path_tidy3d(gf.c.mmi1x2)
    coupler_fdtd = model_from_npz(filepath)
    plot_model(coupler_fdtd)
    plt.show()



from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.array_component import array
from gdsfactory.components.pad import pad
from gdsfactory.components.straight import straight
from gdsfactory.components.via_stack import via_stack as via_stack_factory
from gdsfactory.cross_section import metal2
from gdsfactory.types import ComponentSpec, CrossSectionSpec, Float2


@cell
def array_with_via(
    component: ComponentSpec = pad,
    columns: int = 3,
    spacing: float = 150.0,
    via_spacing: float = 10.0,
    straight_length: float = 60.0,
    cross_section: Optional[CrossSectionSpec] = metal2,
    via_stack: ComponentSpec = via_stack_factory,
    via_stack_dy: float = 0,
    port_orientation: float = 180,
    port_offset: Optional[Float2] = None,
    **kwargs,
) -> Component:
    """Returns an array of vias in X axis with fanout waveguides facing west.

    Args:
        component: to replicate in the array.
        columns: number of components.
        spacing: for the array in um.
        via_spacing: for fanout in um.
        straight_length: length of the straight at the end in um.
        waveguide: waveguide definition.
        cross_section: spec.
        via_stack: spec.
        via_stack_dy: via_stack offset in um.
        port_orientation: 180: facing west.
        port_offset: Optional port movement in um.
        kwargs: cross_section settings.
    """
    c = Component()
    component = gf.get_component(component)
    via_stack = gf.get_component(via_stack)

    for col in range(columns):
        ref = component.ref()
        ref.x = col * spacing
        c.add(ref)

        if port_orientation == 180:
            xlength = col * spacing + straight_length
        elif port_orientation == 0:
            xlength = columns * spacing - (col * spacing) + straight_length
        elif port_orientation == 270:
            xlength = col * via_spacing + straight_length

        elif port_orientation == 90:
            xlength = columns * via_spacing - (col * via_spacing) + straight_length

        else:
            raise ValueError(
                f"Invalid port_orientation = {port_orientation}",
                "180: west, 0: east, 90: north, 270: south",
            )

        via_stack_ref = c << via_stack
        via_stack_ref.x = col * spacing
        via_stack_ref.y = col * via_spacing + via_stack_dy

        if cross_section:
            port_name = f"e{col}"
            straightx_ref = c << straight(
                length=xlength, cross_section=cross_section, **kwargs
            )
            straightx_ref.connect(
                "e2", via_stack_ref.get_ports_list(orientation=port_orientation)[0]
            )
            c.add_port(port_name, port=straightx_ref.ports["e1"])
            if port_offset:
                c.ports[port_name].move(np.array(port_offset) * col)
    return c


@cell
def array_with_via_2d(
    spacing: Float2 = (150.0, 150.0),
    columns: int = 3,
    rows: int = 2,
    **kwargs,
) -> Component:
    """Returns 2D array with fanout waveguides facing west.

    Args:
        spacing: 2D spacing
        columns:
        rows:
        kwargs:
            component: to replicate
            columns: number of components
            spacing: float
            via_spacing: for fanout
            straight_length: length of the straight at the end
            via_stack_port_name:
            **kwargs
    """
    row = array_with_via(columns=columns, spacing=spacing[0], **kwargs)
    return array(component=row, rows=rows, columns=1, spacing=(0, spacing[1]))


if __name__ == "__main__":
    via_stack_big = gf.partial(via_stack_factory, size=(30, 20))
    # c = array_with_via(columns=3, width=10, via_spacing=20, port_orientation=90)
    c = array_with_via_2d(
        columns=2,
        rows=3,
        via_spacing=27.3,
        spacing=(150, 218),
        port_orientation=270,
        straight_length=0,
        via_stack=via_stack_big,
        via_stack_dy=-50 + 10,
        port_offset=(0, 10),
    )
    # c.auto_rename_ports()
    c.show(show_ports=True)



from pathlib import Path
from typing import Union


def check_duplicated_cells(gdspath: Union[Path, str]):
    """Reads cell and checks for duplicated cells.

    Args:
        gdspath: path to GDS or Component

    """
    import klayout.db as pya
    from gdsfactory.component import Component

    if isinstance(gdspath, Component):
        gdspath.flatten()
        gdspath = gdspath.write_gds()
    layout = pya.Layout()
    layout.read(str(gdspath))
    return layout.top_cell()



from typing import Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.array_component import array
from gdsfactory.components.rectangle import rectangle
from gdsfactory.types import Float2, LayerSpec


@cell
def rectangle_with_slits(
    size: Tuple[float, float] = (100.0, 200.0),
    layer: LayerSpec = "WG",
    layer_slit: LayerSpec = "SLAB150",
    centered: bool = False,
    port_type: Optional[str] = None,
    slit_size: Tuple[float, float] = (1.0, 1.0),
    slit_spacing: Float2 = (20, 20),
    slit_enclosure: float = 10,
) -> Component:
    """Returns a rectangle with slits.

    Metal slits reduce stress.

    Args:
        size: (tuple) Width and height of rectangle.
        layer: Specific layer to put polygon geometry on.
        layer_slit: does a boolan NOT when None.
        centered: True sets center to (0, 0), False sets south-west to (0, 0)
        port_type: for the rectangle.
        slit_size: x, y slit size.
        slit_spacing: pitch_x, pitch_y for slits.
        slit_enclosure: from slit to rectangle edge.

    .. code::

        slit_enclosure
        _____________________________________
        |<--->                              |
        |                                   |
        |      ______________________       |
        |     |                      |      |
        |     |                      | slit_size[1]
        |     |______________________|      |
        |  |                                |
        |  | slit_spacing                   |
        |  |                                |  size[1]
        |  |   ______________________       |
        |  |  |                      |      |
        |  |  |                      |      |
        |  |  |______________________|      |
        |     <--------------------->       |
        |            slit_size[0]           |
        |___________________________________|
                        size[0]
    """
    c = Component()
    layer = gf.get_layer(layer)

    r = rectangle(size=size, layer=layer, port_type=port_type, centered=centered)
    c.add_ports(r.ports)
    slit = rectangle(size=slit_size, port_type=None, layer=layer_slit or layer)

    columns = np.floor((size[0] - 2 * slit_enclosure) / slit_spacing[0])
    rows = np.floor((size[1] - 2 * slit_enclosure) / slit_spacing[1])
    slits = array(slit, columns=columns, rows=rows, spacing=slit_spacing).ref()
    slits.xmin = slit_enclosure
    slits.ymin = slit_enclosure

    if layer_slit:
        c << r
        c.add(slits)
    else:
        r_with_slits = c << gf.geometry.boolean(r, slits, operation="not", layer=layer)
        c.absorb(r_with_slits)
    return c


if __name__ == "__main__":
    # c = rectangle_with_slits(layer_slit=None)
    c = rectangle_with_slits(layer_slit=(2, 0), slit_size=(10, 10))
    c.show(show_ports=True)



import gdsfactory as gf
from gdsfactory.components.mzi import mzi as mzi_function
from gdsfactory.components.pad import pad as pad_function
from gdsfactory.components.straight_heater_metal import straight_heater_metal
from gdsfactory.cross_section import metal3, strip
from gdsfactory.routing.get_route import get_route
from gdsfactory.types import ComponentSpec, CrossSectionSpec


@gf.cell
def mzi_pads_center(
    ps_top: ComponentSpec = straight_heater_metal,
    ps_bot: ComponentSpec = straight_heater_metal,
    mzi: ComponentSpec = mzi_function,
    pad: ComponentSpec = pad_function,
    length_x: float = 500,
    length_y: float = 40,
    mzi_sig_top: str = "e3",
    mzi_gnd_top: str = "e2",
    mzi_sig_bot: str = "e1",
    mzi_gnd_bot: str = "e4",
    pad_sig_bot: str = "e1_1_1",
    pad_sig_top: str = "e3_1_3",
    pad_gnd_bot: str = "e4_1_2",
    pad_gnd_top: str = "e2_1_2",
    delta_length: float = 40.0,
    end_straight_length: float = 5,
    start_straight_length: float = 5,
    metal_route_width: float = 10,
    cross_section: CrossSectionSpec = strip,
    cross_section_metal: ComponentSpec = metal3,
) -> gf.Component:
    """Return Mzi phase shifter with pads in the middle.

    GND is the middle pad
    and is shared between top and bottom phase shifters.

    Args:
        ps_top: phase shifter top.
        ps_bot: phase shifter bottom.
        mzi: interforemeter.
        pad: pad function.
        length_x: horizontal length.
        length_y: vertical length.
        mzi_sig_top: port name for top phase shifter signal.
        mzi_gnd_top: port name for top phase shifter GND.
        mzi_sig_bot: port name for top phase shifter signal.
        mzi_gnd_bot: port name for top phase shifter GND.
        pad_sig_bot: port name for top pad.
        pad_sig_top: port name for top pad.
        pad_gnd_bot: port name for top pad.
        pad_gnd_top: port name for top pad.
        delta_length: mzi length imbalance.
        end_straight_length: for routing metal.
        start_straight_length: for routing metal.
        metal_route_width: for routing metal.
        cross_section: for the mzi.
        cross_section_metal: for routing metal.
    """
    c = gf.Component()

    mzi_ps = mzi(
        length_x=length_x,
        straight_x_top=ps_top,
        straight_x_bot=ps_bot,
        length_y=length_y,
        delta_length=delta_length,
        cross_section=cross_section,
    )

    port_names = list(mzi_ps.ports.keys())
    for port_name in [mzi_sig_top, mzi_gnd_top, mzi_sig_bot, mzi_gnd_bot]:
        if port_name not in port_names:
            raise ValueError(f"port {port_name!r} not in {port_names}")

    m = c << mzi_ps
    pads = c << gf.components.array(component=pad, columns=3, rows=1)
    pads.x = m.x
    pads.y = m.y

    route_sig_bot = get_route(
        m.ports[mzi_sig_bot],
        pads.ports[pad_sig_bot],
        cross_section=cross_section_metal,
        bend=gf.components.wire_corner,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        width=metal_route_width,
    )
    c.add(route_sig_bot.references)

    route_gnd_bot = get_route(
        m.ports[mzi_gnd_bot],
        pads.ports[pad_gnd_bot],
        cross_section=cross_section_metal,
        bend=gf.components.wire_corner,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        width=metal_route_width,
    )
    c.add(route_gnd_bot.references)
    route_gnd_top = get_route(
        m.ports[mzi_gnd_top],
        pads.ports[pad_gnd_top],
        cross_section=cross_section_metal,
        bend=gf.components.wire_corner,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        width=metal_route_width,
    )
    c.add(route_gnd_top.references)

    route_sig_top = get_route(
        m.ports[mzi_sig_top],
        pads.ports[pad_sig_top],
        cross_section=cross_section_metal,
        bend=gf.components.wire_corner,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        width=metal_route_width,
    )
    c.add(route_sig_top.references)
    c.add_ports(m.ports)
    return c


if __name__ == "__main__":
    mzi_ps_fa = gf.compose(gf.routing.add_fiber_array, mzi_pads_center)
    mzi_ps_fs = gf.compose(gf.routing.add_fiber_single, mzi_pads_center)
    # c = mzi_ps_fs()
    c = mzi_pads_center()
    c.show(show_ports=True)



"""Design For Testing module includes test protocols."""

from gdsfactory.labels import ehva, siepic, write_labels
from gdsfactory.labels.add_label_yaml import add_label_yaml
from gdsfactory.labels.ehva import DFT, Dft, add_label_ehva
from gdsfactory.labels.merge_test_metadata import merge_test_metadata
from gdsfactory.labels.siepic import add_fiber_array_siepic

__all__ = [
    "DFT",
    "Dft",
    "add_fiber_array_siepic",
    "add_label_ehva",
    "add_label_yaml",
    "ehva",
    "siepic",
    "write_labels",
    "merge_test_metadata",
]



from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.mzit import mzit
from gdsfactory.types import ComponentSpec


@gf.cell
def mzit_lattice(
    coupler_lengths: Tuple[float, ...] = (10.0, 20.0),
    coupler_gaps: Tuple[float, ...] = (0.2, 0.3),
    delta_lengths: Tuple[float, ...] = (10.0,),
    mzi: ComponentSpec = mzit,
) -> Component:
    r"""Mzi fab tolerant lattice filter.

    .. code::

                           cp1
           W3  W1 __                  __ E1___w0_t2   _w2___
                    \                /                      \
                     \    length1   /                        |
                      ============== gap1                    |
                     /              \                        |
                  __/                \_____w0___t1   _w1     |
           W2  W0                       E0               \   | .
                            ...                          |   | .
           W1  W1                                        |   | .
                  __                  __w0____t1____w1___/   |
                    \                /                       |
                     \    lengthN   /                        |
                      ============== gapN                    |
                     /               \                       |
                  __/                 \_                     |
           W0  W0                      \ E0_w0__t2 __w1_____/
                           cpN


    """
    if len(coupler_lengths) != len(coupler_gaps):
        raise ValueError(
            f"Got {len(coupler_lengths)} coupler_lengths and "
            f"{len(coupler_gaps)} coupler_gaps"
        )
    if len(coupler_lengths) != len(delta_lengths) + 1:
        raise ValueError(
            f"Got {len(coupler_lengths)} coupler_lengths and "
            f"{len(delta_lengths)} delta_lengths. "
            "You need one more coupler_length than delta_lengths "
        )

    assert len(coupler_lengths) >= 2

    c = Component()

    cp1 = coupler0 = c << gf.get_component(
        mzi,
        coupler_gap1=coupler_gaps[0],
        coupler_gap2=coupler_gaps[1],
        coupler_length1=coupler_lengths[0],
        coupler_length2=coupler_lengths[1],
        delta_length=delta_lengths[0],
    )

    couplers = [
        c
        << gf.get_component(
            mzi,
            coupler_gap2=coupler_gap,
            coupler_length2=coupler_length,
            coupler1=None,
            delta_length=delta_length,
        )
        for coupler_length, coupler_gap, delta_length in zip(
            coupler_lengths[2:], coupler_gaps[2:], delta_lengths[1:]
        )
    ]

    for i, coupler in enumerate(couplers):
        if i % 2 == 0:
            coupler.mirror()
        coupler.connect("o3", coupler0.ports["o1"])
        coupler.connect("o4", coupler0.ports["o2"])
        coupler0 = coupler

    c.add_port("o1", port=coupler0.ports["o1"])
    c.add_port("o2", port=coupler0.ports["o2"])
    c.add_port("o3", port=cp1.ports["o3"])
    c.add_port("o4", port=cp1.ports["o4"])
    return c


if __name__ == "__main__":
    # cpl = [10, 20, 30]
    # cpg = [0.2, 0.3, 0.5]
    # dl0 = [10, 20]

    cpl = [10, 20, 30, 40]
    cpg = [0.2, 0.3, 0.5, 0.5]
    dl0 = [10, 20, 30]

    c = mzit_lattice(coupler_lengths=cpl, coupler_gaps=cpg, delta_lengths=dl0)
    # c = mzit_lattice()
    c.show(show_ports=True)



from simphony.libraries import sipann


def coupler_ring(
    radius: float = 5.0,
    width: float = 0.5,
    thickness: float = 0.22,
    gap: float = 0.22,
    length_x: float = 4.0,
    sw_angle: float = 90.0,
    **kwargs
):
    r"""Return model for for half a ring coupler.

    Args:
        radius: 5
        width: width um (Valid for 0.4-0.6)
        thickness: Thickness in um (Valid for 0.18-0.24)
        gap: distance between straights in um. (Must be > 0.1)
        length_x: Length of straight portion of coupler in um
        sw_angle: Sidewall angle from horizontal in degrees
        kwargs: geometrical args that this model ignores

    .. code::

           o2 \           / o3
               \         /
                ---------
           o1---------------o4
                 length_x

    .. plot::
        :include-source:

        import gdsfactory as gf
        c = gf.components.coupler_ring()
        c.plot()


    .. plot::
        :include-source:

        import gdsfactory.simulation.simphony as gs
        import gdsfactory.simulation.simphony.components as gc

        m = gc.coupler_ring()
        gs.plot_model(m)

    """
    width *= 1e-6
    thickness *= 1e-6
    gap *= 1e-6
    length = length_x * 1e-6
    radius *= 1e-6

    model = sipann.HalfRacetrack(
        radius=radius,
        width=width,
        thickness=thickness,
        gap=gap,
        length=length,
        sw_angle=sw_angle,
    )
    model.rename_pins("o1", "o2", "o4", "o3")

    return model


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    from gdsfactory.simulation.simphony.plot_model import plot_model

    c = coupler_ring()
    print(c)
    wavelengths = np.linspace(1.5, 1.6) * 1e-6
    plot_model(c, wavelengths=wavelengths)
    plt.show()



"""Route for electrical based on phidl.routing.route_quad."""
from typing import Optional

import numpy as np
import phidl.routing as pr

import gdsfactory as gf
from gdsfactory.port import Port


@gf.cell
def route_quad(
    port1: Port,
    port2: Port,
    width1: Optional[float] = None,
    width2: Optional[float] = None,
    layer: gf.types.LayerSpec = "M1",
) -> gf.Component:
    """Routes a basic quadrilateral polygon directly between two ports.

    Args:
        port1: Port to start route.
        port2 : Port objects to end route.
        width1: Width of quadrilateral at ports. If None, uses port widths.
        width2: Width of quadrilateral at ports. If None, uses port widths.
        layer: Layer to put the route on.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component()
        pad1 = c << gf.components.pad(size=(50, 50))
        pad2 = c << gf.components.pad(size=(10, 10))
        pad2.movex(100)
        pad2.movey(50)
        route_gnd = c << gf.routing.route_quad(
            pad1.ports["e2"],
            pad2.ports["e4"],
            width1=None,
            width2=None,
        )
        c.show()
        c.plot()

    """

    def get_port_edges(port, width):
        _, e1 = pr._get_rotated_basis(port.orientation)
        pt1 = port.center + e1 * width / 2
        pt2 = port.center - e1 * width / 2
        return pt1, pt2

    if width1 is None:
        width1 = port1.width
    if width2 is None:
        width2 = port2.width
    vertices = np.array(get_port_edges(port1, width1) + get_port_edges(port2, width2))
    center = np.mean(vertices, axis=0)
    displacements = vertices - center
    # sort vertices by angle from center of quadrilateral to make convex polygon
    angles = np.array([np.arctan2(disp[0], disp[1]) for disp in displacements])
    vertices = [vert for _, vert in sorted(zip(angles, vertices), key=lambda x: x[0])]

    component = gf.Component()
    component.add_polygon(points=vertices, layer=layer)
    component.add_port(
        name="e1",
        center=port1.center,
        orientation=port1.orientation + 180,
        width=width1,
        layer=layer,
    )
    component.add_port(
        name="e2",
        center=port2.center,
        orientation=port2.orientation + 180,
        width=width2,
        layer=layer,
    )
    return component


if __name__ == "__main__":
    # c = gf.Component()
    # mzi = c << gf.components.mzi_phase_shifter()
    # pads = c << gf.components.array(component=gf.components.pad, columns=3)
    # pads.ymin = mzi.ymax + 30

    # pads.movex(-pads.size_info.sc[0])
    # mzi.movex(-mzi.size_info.sc[0])

    # route_gnd = c << route_quad(
    #     mzi.ports["e1"],
    #     pads.ports["e4_1_1"],
    #     width1=None,
    #     width2=None,
    # )

    # c.show(show_ports=True)

    c = gf.Component()
    pad1 = c << gf.components.pad(size=(50, 50))
    pad2 = c << gf.components.pad(size=(10, 10))
    pad2.movex(100)
    pad2.movey(50)
    route_gnd = c << route_quad(
        pad1.ports["e2"],
        pad2.ports["e4"],
        width1=None,
        width2=None,
    )
    c.show()



"""gdsfactory tidy3d plugin utils."""

from functools import partial
from typing import Optional

from tidy3d import web

from gdsfactory.config import logger


def print_tasks(status: Optional[str] = None) -> None:
    for task in web.get_tasks():
        if status and task["status"] != status:
            continue
        print(task["status"], task["task_name"], task["task_id"])


def delete_tasks(status: str = "error") -> None:
    """Deletes all tasks with a particular status.

    Args:
        status: error, diverged, preprocess

    """
    for task in web.get_tasks():
        # print(task["status"], task["task_name"], task["task_id"])
        if task["status"] == status:
            task_id = task["task_id"]
            logger.info(f"deleted failed task_id {task_id}")
            web.delete(task_id)


delete_tasks_failed = partial(delete_tasks, status="error")
delete_tasks_preprocess = partial(delete_tasks, status="preprocess")
delete_tasks_uploading = partial(delete_tasks, status="uploading")
delete_tasks_diverged = partial(delete_tasks, status="diverged")


if __name__ == "__main__":
    # delete_tasks_failed()
    # print_tasks()
    delete_tasks_uploading()



import pytest

import gdsfactory as gf


def test_get_netlist_cell_array() -> gf.Component:
    c = gf.components.array(
        gf.components.straight(length=10), spacing=(0, 100), columns=1, rows=5
    )
    n = c.get_netlist()
    assert len(c.ports) == 10
    assert not n["connections"]
    assert len(n["ports"]) == 10
    assert len(n["instances"]) == 5
    return c


def test_get_netlist_cell_array_connecting() -> gf.Component:
    c = gf.components.array(
        gf.components.straight(length=100), spacing=(100, 0), columns=5, rows=1
    )
    with pytest.raises(ValueError):
        # because the component-array has automatic external ports, we assume no internal self-connections
        # we expect a ValueError to be thrown where the serendipitous connections are
        c.get_netlist()

    return c


@gf.cell
def test_get_netlist_simple() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i3 = c.add_ref(gf.components.straight(), "i3")
    i2.connect("o2", i1.ports["o1"])
    i3.movey(-100)
    netlist = c.get_netlist()
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    unconnected_optical_port_warnings = netlist["warnings"]["optical"][
        "unconnected_ports"
    ]
    assert len(unconnected_optical_port_warnings) == 1
    assert len(unconnected_optical_port_warnings[0]["ports"]) == 4
    return c


@gf.cell
def test_get_netlist_promoted() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i3 = c.add_ref(gf.components.straight(), "i3")
    i2.connect("o2", i1.ports["o1"])
    i3.movey(-100)
    c.add_port("t1", port=i1.ports["o2"])
    c.add_port("t2", port=i2.ports["o1"])
    c.add_port("t3", port=i3.ports["o1"])
    c.add_port("t4", port=i3.ports["o2"])
    netlist = c.get_netlist()
    connections = netlist["connections"]
    ports = netlist["ports"]
    expected_ports = {"t1": "i1,o2", "t2": "i2,o1", "t3": "i3,o1", "t4": "i3,o2"}

    assert len(connections) == 1
    assert ports == expected_ports
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    assert "warnings" not in netlist
    return c


@gf.cell
def test_get_netlist_close_enough() -> gf.Component:
    """Move connection 1nm inwards."""
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.movex(0.001)
    netlist = c.get_netlist(tolerance=2)
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_close_enough_fails() -> gf.Component:
    """Move connection 1nm outwards."""
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.movex(-0.001)
    netlist = c.get_netlist(tolerance=1)
    connections = netlist["connections"]
    assert len(connections) == 0
    return c


@gf.cell
def test_get_netlist_close_enough_orthogonal() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.movey(0.001)
    netlist = c.get_netlist(tolerance=2)
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_close_enough_orthogonal_fails() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.movey(0.001)
    netlist = c.get_netlist(tolerance=1)
    connections = netlist["connections"]
    assert len(connections) == 0
    return c


@gf.cell
def test_get_netlist_close_enough_both() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.move((0.001, 0.001))
    netlist = c.get_netlist(tolerance=2)
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_close_enough_rotated() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.rotate(angle=0.01, center="o2")
    netlist = c.get_netlist(tolerance=2)
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_throws_error_bad_rotation() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i2.move("o2", destination=i1.ports["o1"])
    i2.rotate(angle=90, center="o2")
    with pytest.raises(ValueError):
        c.get_netlist(tolerance=2)

    return c


@gf.cell
def test_get_netlist_tiny() -> gf.Component:
    c = gf.Component()
    cc = gf.components.straight(length=0.002)
    i1 = c.add_ref(cc, "i1")
    i2 = c.add_ref(cc, "i2")
    i3 = c.add_ref(cc, "i3")
    i4 = c.add_ref(cc, "i4")

    i2.connect("o2", i1.ports["o1"])
    i3.connect("o2", i2.ports["o1"])
    i4.connect("o2", i3.ports["o1"])

    netlist = c.get_netlist(tolerance=5)
    connections = netlist["connections"]
    assert len(connections) == 3
    # cpairs = list(connections.items())
    # extracted_port_pair = set(cpairs[0])
    # expected_port_pair = {'i2,o2', 'i1,o1'}
    # assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_rotated() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(), "i1")
    i2 = c.add_ref(gf.components.straight(), "i2")
    i1.rotate(35)
    i2.connect("o2", i1.ports["o1"])

    netlist = c.get_netlist()
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,o2", "i1,o1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_electrical_simple() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.wire_straight(), "i1")
    i2 = c.add_ref(gf.components.wire_straight(), "i2")
    i3 = c.add_ref(gf.components.wire_straight(), "i3")
    i2.connect("e2", i1.ports["e1"])
    i3.movey(-100)
    netlist = c.get_netlist()
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,e2", "i1,e1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_electrical_rotated_joint() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.wire_straight(), "i1")
    i2 = c.add_ref(gf.components.wire_straight(), "i2")
    i2.connect("e2", i1.ports["e1"])
    i2.rotate(45, "e2")
    netlist = c.get_netlist()
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,e2", "i1,e1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_electrical_allowable_offset() -> gf.Component:
    c = gf.Component()
    i1 = c.add_ref(gf.components.wire_straight(), "i1")
    i2 = c.add_ref(gf.components.wire_straight(), "i2")
    i2.connect("e2", i1.ports["e1"])
    i2.move((0.001, 0.001))
    netlist = c.get_netlist()
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,e2", "i1,e1"}
    assert extracted_port_pair == expected_port_pair
    return c


@gf.cell
def test_get_netlist_electrical_different_widths() -> gf.Component:
    """Move connection 1nm inwards."""
    c = gf.Component()
    i1 = c.add_ref(gf.components.straight(width=1, cross_section="metal1"), "i1")
    i2 = c.add_ref(gf.components.straight(width=10, cross_section="metal1"), "i2")
    i2.move("e2", destination=i1.ports["e1"])
    i2.movex(0.001)
    netlist = c.get_netlist(tolerance=2)
    connections = netlist["connections"]
    assert len(connections) == 1
    cpairs = list(connections.items())
    extracted_port_pair = set(cpairs[0])
    expected_port_pair = {"i2,e2", "i1,e1"}
    assert extracted_port_pair == expected_port_pair
    return c


if __name__ == "__main__":
    # c = gf.c.array()
    # n = c.get_netlist()
    # print(len(n.keys()))
    # c = test_get_netlist_cell_array()
    # c = test_get_netlist_cell_array_connecting()
    # c = test_get_netlist_simple()
    # c = test_get_netlist_promoted()
    # c = test_get_netlist_close_enough()
    # c = test_get_netlist_close_enough_orthogonal()
    # c = test_get_netlist_close_enough_fails()
    # c = test_get_netlist_close_enough_orthogonal_fails()
    # c = test_get_netlist_close_enough_both()
    # c = test_get_netlist_close_enough_rotated()
    # c = test_get_netlist_throws_error_bad_rotation()
    # c = test_get_netlist_tiny()
    # c = test_get_netlist_metal()
    c = test_get_netlist_electrical_different_widths()
    c.show()



from typing import Tuple

from gdsfactory.component import Component
from gdsfactory.types import ComponentOrPath


def check_space(
    gdspath: ComponentOrPath,
    layer: Tuple[int, int] = (1, 0),
    min_space: float = 0.150,
    dbu: float = 1e3,
    ignore_angle_deg: int = 80,
    whole_edges: bool = False,
    metrics: str = "Square",
    min_projection: None = None,
    max_projection: None = None,
) -> int:
    """Reads layer from top cell and returns the area that violates min space.

    If "whole_edges" is true, the resulting EdgePairs collection will receive the whole edges which contribute in the space check.

    "metrics" can be one of the constants Euclidean, Square or Projection. See there for a description of these constants.
    Use nil for this value to select the default (Euclidean metrics).

    "ignore_angle" specifies the angle limit of two edges. If two edges form an angle equal or above the given value, they will not contribute in
    the check. Setting this value to 90 (the default) will exclude edges with an angle of 90 degree or more from the check.
    Use nil for this value to select the default.

    "min_projection" and "max_projection" allow selecting edges by their projected value upon each other. It is sufficient if the projection of on
    e edge on the other matches the specified condition. The projected length must be larger or equal to "min_projection" and less than "max_proje
    ction". If you don't want to specify one limit, pass nil to the respective value.

    Args:
        gdspath: path to GDS or Component.
        layer: tuple.
        min_space: in um.
        dbu: database units (1000 um/nm).
        ignore_angle_deg: The angle above which no check is performed.
        other: The other region against which to check.
        whole_edges: If true, deliver the whole edges.
        metrics: Specify the metrics type 'Euclidian, square'.
        min_projection: lower threshold of the projected length of one edge onto another.
        max_projection: upper limit of the projected length of one edge onto another.

    """
    import klayout.db as pya

    if isinstance(gdspath, Component):
        gdspath.flatten()
        gdspath = gdspath.write_gds()
    layout = pya.Layout()
    layout.read(str(gdspath))
    cell = layout.top_cell()
    region = pya.Region(cell.begin_shapes_rec(layout.layer(layer[0], layer[1])))

    valid_metrics = ["Square", "Euclidean"]

    if metrics not in valid_metrics:
        raise ValueError("metrics = {metrics!r} not in {valid_metrics}")
    metrics = getattr(pya.Region, metrics)

    d = region.space_check(
        min_space * dbu,
        whole_edges,
        metrics,
        ignore_angle_deg,
        min_projection,
        max_projection,
    )
    # print(d.polygons().area())
    return d.polygons().area()


if __name__ == "__main__":
    import gdsfactory as gf
    import klayout.db as pya

    space = 0.12
    min_space = 0.1
    dbu = 1000
    layer = gf.LAYER.WG
    gdspath = gf.components.straight_array(spacing=space)
    gf.show(gdspath)

    if isinstance(gdspath, Component):
        gdspath.flatten()
        gdspath = gdspath.write_gds()

    layout = pya.Layout()
    layout.read(str(gdspath))
    cell = layout.top_cell()
    region = pya.Region(cell.begin_shapes_rec(layout.layer(layer[0], layer[1])))
    print(region.corners().area())
    metrics = "Square"
    metrics = getattr(pya.Region, metrics)
    d = region.space_check(min_space * dbu, False, metrics, 80, None, None)



import numpy as np
from numpy import cos, pi, sin

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.types import LayerSpec


@gf.cell
def ring(
    radius: float = 10.0,
    width: float = 0.5,
    angle_resolution: float = 2.5,
    layer: LayerSpec = "WG",
) -> Component:
    """Returns a ring.

    Args:
        radius: ring radius.
        width: of the ring.
        angle_resolution: number of points per degree.
        layer: layer.
    """
    D = gf.Component()
    inner_radius = radius - width / 2
    outer_radius = radius + width / 2
    n = int(np.round(360 / angle_resolution))
    t = np.linspace(0, 360, n + 1) * pi / 180
    inner_points_x = (inner_radius * cos(t)).tolist()
    inner_points_y = (inner_radius * sin(t)).tolist()
    outer_points_x = (outer_radius * cos(t)).tolist()
    outer_points_y = (outer_radius * sin(t)).tolist()
    xpts = inner_points_x + outer_points_x[::-1]
    ypts = inner_points_y + outer_points_y[::-1]
    D.add_polygon(points=(xpts, ypts), layer=layer)
    return D


if __name__ == "__main__":
    c = ring(radius=5)
    c.show(show_ports=True)
